/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: StandardConnectionAllocator.java,v 1.2 2003/05/19 16:27:14 wiznogood Exp $
 */

package	com.lutris.appserver.server.sql.standard;

import java.sql.SQLException;
import java.util.Date;
import java.util.Stack;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.enhydra.dods.DODS;

import com.lutris.appserver.server.sql.ConnectionAllocator;
import com.lutris.appserver.server.sql.DBConnection;
import com.lutris.appserver.server.sql.LogicalDatabase;
import com.lutris.util.Config;
import com.lutris.util.ConfigException;
import com.lutris.util.KeywordValueException;


/**
 * Manages a pool (set) of connections to a database.  The pool is named
 * as a logical database.  By naming a pool, this allows connection  resource
 * control to be on a finer grain that a database and allows for easier
 * migration to multiple databases.  One or more pools can map to the same
 * actual database.  A connection considered part of the pool, even if its
 * allocated to a thread.  These objects are all publicly accessed via the
 * Database Manager, not directly.
 * <P>
 * If an error occurs in a connection, it is dropped from the pool.  The
 * process using the connection has already received an error which aborts
 * the work in progress.  By dropping the connection, waiting threads are
 * restarted.  If something is wrong with the database (e.g. server is down),
 * they will recieve errors and also be aborted.  A generation number is used
 * to close down all connections that were open when the error occured,
 * allowing new connections to be allocated.
 * <P>
 * The configuration data is specified in the section:
 * <B><CODE>DatabaseManager.DB.<I>dbName</I>.Connection</CODE></B>
 * <P>
 * <I>Configuration sub fields are:</I>
 * <UL>
 * <LI> <B><CODE>Url</CODE></B> -
 *      The JDBC URLof the database. Manditary.
 *      E.g. "jdbc:sequelink://dbHost:4000/[Informix];Database=dummy"
 * <LI> <B><CODE>User</CODE></B> -
 *      The database users used to access the database. Manditary.
 * <LI> <B><CODE>Password</CODE></B> -
 *      The database user's password. Manditary.
 * <LI> <B><CODE>MaxPoolSize</CODE></B> -
 *      The maximum number of open connections to the database. Optional,
 *      if not specified, then it default to 0. A value of 0 means that
 *      connections are allocated indefinitely or until the database (JDBC)
 *      refuses any new connections.
 * <LI> <B><CODE>Logging</CODE></B> -
 *      Specify true to enable SQL logging, false to disable it. Optional,
 *      false if not specified.
 * <LI> <B><CODE>AllocationTimeout</CODE></B> -
 *      The Maximum amount of time that a thread will wait for
 *      a connection from the connection allocator before an exception is
 *      thrown. This will prevent possible dead locks. The time out is in
 *      milliseconds. If the time out is <= 0, the allocation of connections
 *      will wait indefinitely. Optional, if not specified, then it
 *      defaults to 1000 (ms).
 * <LI> <B><CODE>QueryTimeout</CODE></B> - The amount of time (in seconds) that
 *	a query will block before throwing an exception. If <= 0 then the
 *	query will not block. Optional, if not specified, then the value
 *	defaults to 0. This is not implemented by all logical databases.
 * <LI> <B><CODE>TransactionTimeout</CODE></B> - The amount of time (in seconds)
 *	that a transaction will block before throwing an exception. If
 *	<= 0 then the transaction will not block. Optional, if not specified,
 *	then the value defaults to 0. This is not implemented by all
 *	logical databases.
 * <LI> <B><CODE>MaxPreparedStatements</CODE></B> - If specified, overrides
 *      the JDBC <CODE>Connection.getMetaData().getMaxStatements()</CODE>
 *      value.  If less than zero, use the meta data value.  Optional,
 *      default is to use the meta data.
 * </UL>
 * It would be nice to add a config parameter that would disable caching
 * of PreparedStatements.
 *
 * @author	Mark Diekhans
 * @author	Kyle Clark
 * @author	Paul Morgan
 * @since	LBS1.8
 * @version	$Revision: 1.2 $
 */
public class StandardConnectionAllocator implements ConnectionAllocator {

    /**
     * Reference to the logical database for easy access to the
     * connection pool.
     */
    protected LogicalDatabase logicalDatabase = null;

    /**
     * JDBC URL of database.
     */
    protected String url;

    /**
     * SQL user name.
     */
    protected String user;

    /**
     * SQL password..
     */
    protected String password;

    /**
     * Maximum number of connections in the pool.
     * If this value is <= zero, then create as many
     * connections as possible.
     */
    private int maxPoolSize;

    /**
     * Current size of the pool; includes allocated connections.
     */
    private int currentPoolSize;

    /**
     * Maximum size the pool ever got to, regardless of generation.
     */
    private int biggestPoolSize;

    /**
     * Date at which the biggest pool size occured.
     */
    private Date biggestPoolDate;

    /**
     * Number of queries or transactions on this logical database.
     */
    protected long numRequests;

    /**
     * The actual pool of DBConnection objects.
     */
    private Stack pool;

    /**
     * Indicates if logging is enabled.
     */
    protected boolean sqlLogging;

    /**
     * Maximum amount of time in milliseconds to wait for a connection.
     */
    private int timeOut;

    /**
     * Maximum amount of time in seconds to block on a query. The
     * DBQuery object will retrieve this value from the connection.
     */
    protected int queryTimeOut;

    /**
     * Maximum amount of time in seconds to block on a transaction. The
     * DBTransaction object will retrieve this value from the connection.
     */
    protected int transactionTimeOut;

    /**
     * Maximum number of prepared statements to use; if less-than zero,
     * then JDBC is queried for this value.
     */
    protected int maxPreparedStatements;

    /**
     * Generation number.  When an SQL error occurs, all objects of the
     * same generation or earlier are dropped.
     */
    protected int generation = 1;

    /**
     * The log channel.
     */
    private Logger logger = DODS.getLogger();


    /**
     * Create a new connection in the pool.
     *
     * @exception java.sql.SQLException If a SQL error occures.
     */
    protected DBConnection createConnection ()
    throws java.sql.SQLException {
        DBConnection dbConnection = new StandardDBConnection (this, url, user, 
                                        password, maxPreparedStatements,
				                        sqlLogging, generation);
        return dbConnection;
    }

    /**
     * Initialize the connection allocator object.  Connections are
     * opened on demand and stored in a pool.
     *
     * @param url JDBC URL of database.
     * @param user SQL user name.
     * @param password SQL password.
     * @param maxPoolSize Maximum number of connections.
     * @param sqlLogging Specifying <CODE>true</CODE> enables SQL logging.
     * @param timeOut Set the maximum amount of time in milliseconds
     *  to wait for a new connection.
     * @exception ConfigException if bad configuration information is
     *  given in the config file.
     */
    protected StandardConnectionAllocator (LogicalDatabase logicalDatabase,
                                           Config conConfig)
    throws ConfigException {
    this.logicalDatabase = logicalDatabase;
    
        try {
            url = conConfig.getString ("Url");
            user = conConfig.getString ("User");
            password = conConfig.getString ("Password");
            timeOut = conConfig.getInt("AllocationTimeout", 1000);
            maxPoolSize = conConfig.getInt ("MaxPoolSize", 0);
            sqlLogging = conConfig.getBoolean ("Logging", false);
            queryTimeOut = conConfig.getInt("QueryTimeout", 0);
            transactionTimeOut = conConfig.getInt("TransactionTimeout", 0);
            maxPreparedStatements = conConfig.getInt("MaxPreparedStatements", -1);
        } catch (KeywordValueException except) {
            throw new ConfigException ("Bad DatabaseManager.DB." + logicalDatabase.getName() 
                                       + ".Connection section defined in config file.");
        }
        currentPoolSize = 0;
        pool = new Stack ();
        
        biggestPoolSize = 0;
        biggestPoolDate = new Date();
        numRequests = 0;
    }


    /**
     * Allocate a connection to a thread.  If none are available, grow the
     * pool.  If the pool is alredy its maximum size, then the thread waits.
     *
     * @return The allocated connection object.
     * @exception SQLException
     *   If a SQL error occures.
     */
    public synchronized DBConnection allocate ()
    throws SQLException {
        //
        // It isn't always possible to determine the maximum
        // number of connections allowed to the database because
        // of JDBC driver differences. We assume connections are
        // available until we fail to allocate one or we reach
        // the maximum configured.
        //
        boolean createNewConn = true;
        while (pool.empty()) {
            if (createNewConn && ((currentPoolSize < maxPoolSize) 
                || (maxPoolSize <= 0))) {
                try {
                    pool.push(createConnection());
                    currentPoolSize++;
        		    if (currentPoolSize > biggestPoolSize) {
            			biggestPoolSize = currentPoolSize;
            			biggestPoolDate = new Date();
        		    }
                } catch (SQLException e) {
                    if (currentPoolSize > 0) {
// v. strahinja, 27 sep 2002                          log.write(Logger.NOTICE,
                        logger.log(Level.INFO,
                            "ConnectionAllocator: " +
                            "failed to allocate a new connection due to" +
                            e.toString() +
                            "Error code: " + e.getErrorCode() + "\n" +
                            "SQLState: " + e.getSQLState() + "\n" +
                            "\nCurrent pool size is: " + currentPoolSize +
                            "\nMaximum configured pool size is now " +
                            maxPoolSize + "\nContinuing...\n");
                        createNewConn = false;
                    } else {
// v. strahinja, 27 sep 2002                          log.write(Logger.ALERT,
                        logger.log(Level.FATAL,
                             "ConnectionAllocator: " +
                             "failed to allocate a new connection" +
                             "\nThe connection pool is empty!\n");
                        throw e;
                    }
                }
            } else {
                try {
                    if (timeOut > 0) {
                        wait(timeOut);
                        if (pool.empty()) {
// v. strahinja, 27 sep 2002                              log.write(Logger.ALERT,
                            logger.log(Level.FATAL,
                                 "ConnectionAllocator: " +
                                 "allocation of a new connection timed out." +
                                 "Possible dead lock avoided.");
                            String msg =
                                "Connections are currently unavailable.\n" +
                                "Possible dead lock avoided.";
                            throw new SQLException(msg);
                        }
                    } else {
                        wait();
                    }
                } catch (InterruptedException intEx) {
                }
            }
        }

        //
        // A connection is available.
        //
        DBConnection conn = (DBConnection)pool.pop();
        conn.allocate();
        return conn;
    }


    /**
     * Return a connection to the pool.  If it is of an old generation,
     * close and drop.
     *
     * @param dbConnection The connection object to return.
     */
    public synchronized void release (DBConnection dbConnection) {

        if ((dbConnection.getGeneration () < generation)
	            || dbConnection.isMarkedForDrop()) {
            dbConnection.close ();
            currentPoolSize--;
        } else {
            pool.push (dbConnection);
        }
        notify ();
    }


    /**
     * Called when a connection in this pool has an SQL error.
     * All unallocated connections in the pool are dropped if
     * they have a generation number less
     * than or equal to the error connection.
     * If the current generation number
     * is the same as this generation number, increment it.
     * This way so that all outstanding connections
     * (including the one passed in) of the same generation
     * are dropped when returned.
     *
     * @param dbConnection The connection object to drop.
     *   The connection should be
     *   returned to the pool after this function returns.
     */
    public synchronized void drop (DBConnection dbConnection) {

        if (generation <= dbConnection.getGeneration ()) {
            generation++; // new generation
        }

        // Delete all entries of the last generation.
        Stack holdPool = new Stack ();
        while (!pool.empty ()) {
            DBConnection connect = (DBConnection) pool.pop ();
            if (connect.getGeneration () < generation) {
                connect.close ();
                currentPoolSize--;
            } else {
                holdPool.push (connect);
            }
        }

        // Put all current generation entries back in the pool.
        while (!holdPool.empty ()) {
            pool.push (holdPool.pop ());
        }
        notify ();
    }

    /**
     * Called when the database manager is shutting down:
     * Close all connections immediately.
     */
    public synchronized void dropAllNow () {
        while (!pool.empty ()) {
            DBConnection connect = (DBConnection) pool.pop ();
            connect.close ();
            currentPoolSize--;
	    }
    }

    /**
     * Return the number of currently active connections.
     *
     * @return The number of connections.
     */
    public int getActiveCount () {
    	return currentPoolSize;
    }

    /**
     * Return the maximum number of connections active at one time.
     *
     * @return The number of connections.
     */
    public int getMaxCount () {
    	return biggestPoolSize;
    }

    /**
     * Return the time when the maximum connection count occured.
     *
     * @return The <CODE>Date</CODE> when the maximum connection
     *  count occured.
     */
    public Date getMaxCountDate () {
    	return biggestPoolDate;
    }

    /**
     * Reset the maximum connection count and date.
     */
    public void resetMaxCount () {
    	biggestPoolSize = currentPoolSize;
	    biggestPoolDate = new Date();
    }

    /**
     * Return the number of database requests.
     *
     * @return The number of database requests (queries or transactions).
     */
    public long getRequestCount () {
    	return numRequests;
    }

    /**
     * Finalizer.
     * If any connections allocated by this object have not been closed,
     * this method ensures that garbage collection does so.
     */
    protected void finalize() {
    	dropAllNow();
    }
}