/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 // compliance with WEBDOCWF begin
 * contains WebDocWf bug fix, concurrent applications should be allowed to generate oid's
 // original line
 //
 // compliance with WEBDOCWF end
 * Contributor(s):
 *
 * $Id: StandardObjectIdAllocator.java,v 1.2 2003/05/19 16:27:14 wiznogood Exp $
 *
 // compliance with WEBDOCWF begin
 * contains WebDocWf bug fix, concurrent applications should be allowed to generate oid's
 // original line
 //
 // compliance with WEBDOCWF end
 *
 * formatted with JxBeauty (c) johann.langhofer@nextra.at
 */


package com.lutris.appserver.server.sql.standard;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.enhydra.dods.DODS;

import com.lutris.appserver.server.sql.DBConnection;
import com.lutris.appserver.server.sql.LogicalDatabase;
import com.lutris.appserver.server.sql.ObjectId;
import com.lutris.appserver.server.sql.ObjectIdAllocationError;
import com.lutris.appserver.server.sql.ObjectIdAllocator;
import com.lutris.appserver.server.sql.ObjectIdException;
import com.lutris.util.Config;
import com.lutris.util.ConfigException;
import com.lutris.util.KeywordValueException;

/**
 * Object ids can only be created via this manager.
 * Ensures that all object ids are unique across all
 * objects in this logical database. Also ensures good performance
 * for allocating object ids.
 * <P>
 * The configuration data is specified in the section:
 * <CODE>DatabaseManager.DB.<I>dbName</I>.ObjectId</CODE>
 * <P>
 * Configuration fields are:
 * <UL>
 * <LI> <CODE>CacheSize</CODE> -
 *      The number of object id's to cache between database queries. Optional,
 *	if not specified, then it defaults to 1024.
 * <LI> <CODE>MinValue</CODE> -
 *      The starting number of Object ID allocation. This will only be used
 *	if the Object ID table is empty and thus is useful in development
 *	and testing. Optional, if not specified it defaults to
 *	100000000000000000. Note that the largest number that can be
 *	associated with an OID in LBS is "database: DECIMAL(19,0)"
 * </UL>
 *
 * @author	Kyle Clark
 * @author	Paul Morgan
 * @since	LBS1.8
 * @version	$Revision: 1.2 $
 */
public class StandardObjectIdAllocator implements ObjectIdAllocator {
    
        /**
         * The maximum object id in the cache.
         */
        private ObjectId max = null;
        
        /**
         * The object id mos recently allocated.
         */
        private ObjectId current = null;
        
        /**
         * The next object id that will be allocated.
         */
        private ObjectId next = null;
        
        /**
         * Default cache size if 1K.  Note that if the
         * application is started and stopped frequently
         * you should drop this value in order not
         * to waste object ids.
         */
        private long cacheSize = 1024;
        
        /**
         * Starting value for object id creation, if table is empty.
         * This is a string because very large numbers are possible.
         */
        private String oidMinValue = "0";
        
        /**
         * Reference to the logical database for easy access to the
         * connection pool.
         */
        private LogicalDatabase logicalDatabase = null;
    
        /**
         * If allocation of object id's fails (because the contents of
         * the table are modified by another process) this object
         * id manager will keep trying to allocate object ids
         * until it succeeds.  If it fails more than LOG_THRESHOLD
         * times, then it will log a message to the system.
         */
        private static final int LOG_THRESHOLD = 100;
    
        /**
         * The log channel.
         */
        private Logger logger = DODS.getLogger();
    
        /**
         * Initialize the object id manager.
         *
         * @param objCacheSize The number of object ids to cache between
         *  database queries.
         * @param objIdMinValue The starting number for object ids. Used
         *  only if object id table is empty.
         * @exception ConfigException if bad configuration information is
         *  given in the config file.
         */
        protected StandardObjectIdAllocator (LogicalDatabase logicalDatabase, Config objIdConfig) 
        throws ConfigException {
            this.logicalDatabase = logicalDatabase;
            try {
                cacheSize = objIdConfig.getInt("CacheSize", 1024);
                oidMinValue = objIdConfig.getString("MinValue", "10000000000000000");
            } catch (KeywordValueException except) {
                throw  new ConfigException("Bad DatabaseManager.DB." + logicalDatabase.getName()
                + ".ObjectId section defined in config file.");
            }
        }
    
        /**
         * Check if oid belongs to Object id's range [minOId, currentOId]
    	 *
         * @param oid
         *   oid which will be checked.
         * @exception ObjectIdException
         *   If a oid does't belong to range.
         */
        public synchronized void checkOId(ObjectId oid)  throws ObjectIdException {
    		if (oid == null) 
    			throw new ObjectIdException("Object id is null");
    		BigDecimal min = new BigDecimal(new BigInteger(oidMinValue));
    		BigDecimal id = new BigDecimal(oid.toString());
    		if (id.compareTo(min) == -1) {
    			throw new ObjectIdException("Object id out of range");
    		}
    		if (next == null || max == null) {
    			updateCache();  
    		}
    		BigDecimal bdNext = new BigDecimal(next.toString());
    		if (id.compareTo(bdNext) == 0 || id.compareTo(bdNext) == 1) {
    			throw new ObjectIdException("Object id out of range");
    		}
    	}
    
        /**
         * Allocates a new object id.
         */
        public synchronized ObjectId allocate () {
            try {
                if (next == null || max == null || next.equals(max)) {
                    updateCache();
                }
                current = next; // next free
                next = next.increment(); 
                return  current;
            } catch (Exception e) {
                throw  new ObjectIdAllocationError("ObjectIdAllocator: " + "\nFailed to allocate object id.  Caused by: "
                                                   + "\n" + e.getMessage());
            }
        }
    
    /**
    * Updates the cache of object id's.
    *
    * contains WebDocWf bug fix, concurrent applications should be allowed to generate oid's
    */
    private void updateCache () {
        final String OID_TABLE = "objectid";
        DBConnection conn = null;
        try {
            boolean tryAgain = false;
            int tryCount = 0;
            conn = logicalDatabase.allocateConnection();
            conn.setAutoCommit(false);
            do {
                BigDecimal dbNext;
                ResultSet rs;
                // Query next available object id from database.
                rs = conn.executeQuery("select * from " + OID_TABLE);
                if (!rs.next()) {
                    // Table is empty - initialize
                    conn.executeUpdate("insert into " + OID_TABLE + " values(" + oidMinValue + ")");
                    dbNext = new BigDecimal(oidMinValue);
                } else {
                    dbNext = rs.getBigDecimal("next", 0);
                }
                rs.close();             // In case there's more // JOHN
                // Sync this managers next oid w/ current value in db.
                next = new ObjectId(dbNext);
                // Update db contents
                // i.e. Allocate a block of oids of size cacheSize.
                BigDecimal dbLast = dbNext;
                dbNext = dbNext.add(BigDecimal.valueOf(cacheSize));
                // This now updates all rows if there's more than one.
                String sql = "update " + OID_TABLE + " \n" +
                // compliance with WEBDOCWF begin
                // WebDocWf bug fix, concurrent applications should be allowed to generate oid's
                // the "+ "where next = ?"" has been under comments before
                "set next = ?" + " \n" + "where next = ?";
                // end of first part of bugfix
                // original line
                // compliance with WEBDOCWF end
                PreparedStatement stmt = conn.prepareStatement(sql);
                stmt.setBigDecimal(1, dbNext);
                // compliance with WEBDOCWF begin
                // WebDocWf bug fix, concurrent applications should be allowed to generate oid's
                // The following line has been under comments before:
                stmt.setBigDecimal(2, dbLast);
                // end of second part of bugfix
                // original line
                // compliance with WEBDOCWF end
                if (conn.executeUpdate(stmt, sql) < 1) {
                    // We were unable to update the table.
                    tryAgain = true;
                    tryCount++;
                    if (tryCount >= LOG_THRESHOLD) {
                        // v. strahinja, 27 sep 2002                          log.write(Logger.ALERT,
                        logger.log(Level.FATAL, "ObjectIdAllocator: " + "\n  Failed to allocate object ids."
                        + "\n  Trying again....");
                    }
                    // Transaction rollback
                    conn.rollback();
                    conn.reset();         // free resources
                    //conn.setLockModeToWait(true);
                    conn.setAutoCommit(false);
                    if (tryCount >= 50) {
                        // v. strahinja, 27 sep 2002                          log.write(Logger.ALERT,
                        logger.log(Level.FATAL, "ObjectIdAllocator: " + "\n  Failed to allocate object ids."
                        + "\n  Tried 50 times.  Giving up.");
                        throw  new ObjectIdAllocationError("Failed to allocate object id."
                        + "\nTried 50 times.  Giving up.");
                    }
                } else {
                    tryAgain = false;
                    // Sync this managers max oid w/ db contents.
                    max = new ObjectId(dbNext);
                    // Transcation commit.
                    conn.commit();
                    if (tryCount >= LOG_THRESHOLD) {
                        // v. strahinja, 27 sep 2002                          log.write(Logger.ALERT,
                        logger.log(Level.FATAL, "ObjectIdAllocator: " + "\n  Succeeded in allocating object ids."
                        + "\n  Continuing...");
                        }
                }
            } while (tryAgain);
        } catch (SQLException sqlExcept) {
            // An error occured.  Rollback all changes.
            max = next;               // The cache is not updated.
            // v. strahinja, 27 sep 2002              log.write(Logger.ALERT,
            logger.log(Level.FATAL, "ObjectIdAllocator: " + "\n  Failed to allocate object ids.  Giving up!"
                        + "\n    SQLState = " + sqlExcept.getSQLState() + "\n    SQLError = "
                        + sqlExcept.getErrorCode() + "\n    SQLMsg   = " + sqlExcept.getMessage());
            if (conn != null) {
                try {
                    conn.rollback();
                } catch (SQLException rollbackExcept) {
                    // v. strahinja, 27 sep 2002                      log.write(Logger.ALERT,
                    logger.log(Level.FATAL, "ObjectIdAllocator: " + "\n  Failed to rollback transaction."
                                + "\n    SQLState = " + rollbackExcept.getSQLState() + "\n    SQLError = "
                                + rollbackExcept.getErrorCode() + "\n    SQLMsg   = " + rollbackExcept.getMessage());
                }
                if (!conn.handleException(sqlExcept))
                    conn = null;
            } throw  new ObjectIdAllocationError("ObjectIdAllocator: " + "\n  Failed to allocate object id.  Caused by SQLException:"
                    + "\n    SQLState = " + sqlExcept.getSQLState() + "\n    SQLError = "
                    + sqlExcept.getErrorCode() + "\n    SQLMsg   = " + sqlExcept.getMessage());
        } catch (ObjectIdException oidEx) {
            throw  new ObjectIdAllocationError("ObjectIdAllocator: " + "\nFailed to allocate object id.  Caused by: "
            + "\n" + oidEx.getMessage());
        } finally {
            if (conn != null) {
                try {
                    conn.reset();         // free existing resource if any
                } catch (SQLException sqlExcept) {
                    // v. strahinja, 27 sep 2002                      log.write(Logger.ALERT,
                    logger.log(Level.FATAL, "ObjectIdAllocator: " + "\n  Failed to reset connection."
                    + "\n    SQLState = " + sqlExcept.getSQLState() + "\n    SQLError = "
                    + sqlExcept.getErrorCode() + "\n    SQLMsg   = " + sqlExcept.getMessage());
                } finally {
                    conn.release();
                }
            }
        }
    }
}