
/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 */

/*
 *
 * @author    Nenad Vico
 * @author    Tanja Jovanovic
 * @version   1.0.0  2003/03/16
 *
 */

package org.enhydra.dods;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import com.lutris.appserver.server.sql.DatabaseManager;
import com.lutris.appserver.server.sql.DatabaseManagerException;
import com.lutris.appserver.server.sql.StandardDatabaseManager;
import com.lutris.util.ConfigException;

/**
 * Main DODS class.
 *
 * There are two modes of usage: non-threading and threading.
 * In non-threading mode, only one DatabaseManager is used for the whole
 * application, no matter the application has one or more Threads.
 * In threading mode, there is one DatabaseManager for every Thread.
 * User needs, for every Thread, to define the DatabaseManager.
 * If, for any Thread, the DatabaseManager is not defined,
 * the default DatabaseManager is used.
 *
 * <p>
 * Example for non-threading mode:
 * <blockquote><pre>
 *  ...
 *  try {
        String fileName = "discRack.conf";
 *      DatabaseManager dbManager = StandardDatabaseManager.newInstance(fileName);
 *      DODS.register(dbManager);
 *  } catch (Exception e) {
 *      e.printStackTrace();
 *  }
 *  ...
 * </pre></blockquote>
 */

public class DODS {

    protected static final String LOG4J_CONF_FILE = "log4j.xml";

    /**
     * This attribute is used for storing <code>DatabaseManager</code>s, if
     * threading mode is used.
     * Every <code>Thread</code> has precisely one <code>DatabaseManager</code>.
     * If <code>Thread</code> doesn't have <code>DatabaseManager</code>, it uses
     * <code>defaultDatabaseManager</code>.
     */
    private static HashMap databaseManagers;

    /**
     * This attribute is used for storing default <code>DatabaseManager</code>.
     * In non-treading mode, there is only one <code>DatabaseManager</code>, and
     * this <code>DatabaseManager</code> is set as <code>defaultDatabaseManager</code>.
     */
    private static DatabaseManager defaultDatabaseManager;

    /**
     * This attribute is used for storing <code>Logger</code>s, if
     * threading mode is used.
     * Every <code>Thread</code> has precisely one <code>Logger</code>.
     * If <code>Thread</code> doesn't have <code>Logger</code>, it uses
     * <code>defaultLogger</code>.
     */
    private static HashMap loggers;

    /**
     * This attribute is used for storing default <code>Logger</code>.
     * <code>defaultLogger</code> is initialized with the root <code>Logger</code>.
     * It is assumed that user sets root <code>Logger</code> in log4j configuration.
     * Otherwise, logging is set to OFF.
     */
    private static Logger defaultLogger;

    /**
     * Mode of usage. If this flag is <code>true</code>, DODS uses multi
     * threading mode. This means that DODS handles multiple threads, like
     * Enhydra application server (each application in one <code>Thread</code>).
     * If this flag is <code>false</code>, there can be more then one <code>Thread</code>,
     * but there is only one <code>DatabaseManager</code> for all <code>Thread</code>s
     * (for the whole application).
     */
    private static boolean threading;

    static {
        init();
    }

    /**
     * Sets the default DatabaseManager.
     *
     * @param databaseManager the DatabaseManager to be set as default.
     */
    public static void registerDefault(DatabaseManager databaseManager) {
        defaultDatabaseManager = databaseManager;
    }

    /**
     * Sets the default DatabaseManager.
     *
     * @param fileName full path name of the application configuration file.
     */
    public static void registerDefault(String fileName)
    throws ConfigException, DatabaseManagerException, SQLException {
        defaultDatabaseManager = StandardDatabaseManager.newInstance(fileName);
    }

    /**
     * In threading mode, this method associates the DatabaseManager object
     * with the current thread.
     * If default DatabaseManager is <code>null</code>, it will also be set.
     * In non-threading mode, only default DatabaseManager will be set.
     *
     * @param databaseManager the DatabaseManager to associate
     *        with current thread.
     */
    public static void register(DatabaseManager databaseManager) {
        if (threading) {
            databaseManagers.put(Thread.currentThread(),databaseManager);
            if (defaultDatabaseManager == null)
                defaultDatabaseManager = databaseManager;
            return;
        }
        defaultDatabaseManager = databaseManager;
    }

    /**
     * In threading mode, this method associates the DatabaseManager object
     * with the current thread.
     * If default DatabaseManager is <code>null</code>, it will also be set.
     * In non-threading mode, only default DatabaseManager will be set.
     *
     * @param fileName full path name of the application configuration file
     *      which will create <code>DatabaseManager</code>associate with current thread.
     */
    public static void stratup(String fileName)
    throws ConfigException, DatabaseManagerException, SQLException {
        DatabaseManager databaseManager = StandardDatabaseManager.newInstance(fileName);
        if (threading) {
            databaseManagers.put(Thread.currentThread(),databaseManager);
            if (defaultDatabaseManager == null)
                defaultDatabaseManager = databaseManager;
            return;
        }
        defaultDatabaseManager = databaseManager;
    }

    /**
     * Associates the given DatabaseManager object with given thread.
     * If default DatabaseManager is null, the default DatabaseManager
     * will also be set.
     *
     * @param thread the thread to associate with <code>DatabaseManager</code>.
     * @param databaseManager the DatabaseManager to associate the <code>Thread</code>.
     */
    public static void register(Thread thread, DatabaseManager databaseManager) {
        databaseManagers.put(thread,databaseManager);
        if (defaultDatabaseManager == null)
            defaultDatabaseManager = databaseManager;
    }

    /**
     * Associates DatabaseManager created by the given fileName with given thread.
     * If default DatabaseManager is null, the default DatabaseManager
     * will also be set.
     *
     * @param thread the thread to associate with <code>DatabaseManager</code>.
     * @param fileName full path name of the application configuration file
     *      which will create <code>DatabaseManager</code>associate with given thread.
     */
    public static void startup(Thread thread, String fileName)
    throws ConfigException, DatabaseManagerException, SQLException {
        DatabaseManager databaseManager = StandardDatabaseManager.newInstance(fileName);
        databaseManagers.put(thread,databaseManager);
        if (defaultDatabaseManager == null)
            defaultDatabaseManager = databaseManager;
    }

   /**
     * Sets the default logger.
     *
     * @param name logger name, set in log4j configuration file.
     */
    public static void registerDefaultLogger(String name) {
        defaultLogger = Logger.getLogger(name);
    }

    /**
     * Sets the default logger.
     *
     * @param logger logger that will be set as default logger.
     */
    public static void registerDefaultLogger(Logger logger) {
        defaultLogger = logger;
    }

    /**
     * In threading mode, this method associates the logger object with
     * the current thread.
     * In non-threading mode, only default logger will be set.
     *
     * @param name logger name, set in log4j configuration file.
     */
    public static void registerLogger(String name) {
        if (threading) {
            loggers.put(Thread.currentThread(),Logger.getLogger(name));
            return;
        }
        defaultLogger = Logger.getLogger(name);
    }

    /**
     * In threading mode, this method associates the <code>logger</code> object with
     * the current thread.
     * In non-threading mode, only default <code>logger</code> will be set.
     *
     * @param logger logger that will be set.
     */
    public static void registerLogger(Logger logger) {
        if (threading) {
            loggers.put(Thread.currentThread(),logger);
            return;
        }
        defaultLogger = logger;
    }

    /**
     * Associates the given logger object with given thread.
     *
     * @param thread the thread to associate with the logger.
     * @param name the logger to associate the <code>Thread</code>.
     */
    public static void registerLogger(Thread thread, String name) {
        loggers.put(thread,Logger.getLogger(name));
    }

    /**
     * Associates the given <code>logger</code> object with the given thread.
     *
     * @param thread the thread to associate with <code>logger</code>.
     * @param logger the <code>logger</code> to associate the <code>Thread</code>.
     */
    public static void registerLogger(Thread thread, Logger logger) {
        loggers.put(thread,logger);
    }

    /**
     * Unregisters default <code>DatabaseManager</code>.
     * Call this method to release default <code>DatabaseManager</code>.
     *
     * @return unregistered default <code>DatabaseManager</code>.
     *
     * @exception DODSException
     *   If an error occurs in unregistering the DatabaseManager.
     */
    public static DatabaseManager unregisterDefault()
    throws DODSException {
        try {
            DatabaseManager dbManager = defaultDatabaseManager;
            defaultDatabaseManager = null;
            return dbManager;
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }

    /**
     * In threading mode, this method unregisters <code>DatabaseManager</code>
     * associated with the current thread.
     * In non-threading mode, only default <code>DatabaseManager</code>
     * will be unregistered.
     * Call this method to release <code>DatabaseManager</code>.
     *
     * @return unregistered <code>DatabaseManager</code>.
     *
     * @exception DODSException
     *   If an error occurs in unregistering the DatabaseManager.
     */
    public static DatabaseManager unregister()
    throws DODSException {
        try {
            DatabaseManager dbManager;
            if (threading)
                return (DatabaseManager)databaseManagers.remove(Thread.currentThread());
            dbManager = defaultDatabaseManager;
            defaultDatabaseManager = null;
            return dbManager;
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }

    /**
     * Unregisters <code>DatabaseManager</code> associated with the given thread.
     * Call this method to release <code>DatabaseManager</code>.
     *
     * @return unregistered <code>DatabaseManager</code>.
     *
     * @exception DODSException
     *   If an error occurs in unregistering the DatabaseManager.
     */
    public static DatabaseManager unregister(Thread thread)
    throws DODSException {
        try {
            return (DatabaseManager)databaseManagers.remove(thread);
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }

    /**
     * Unregisters default<code>Logger</code>.
     * Call this method to release default <code>Logger</code>.
     *
     * @return unregistered default <code>Logger</code>.
     *
     * @exception DODSException
     *   If an error occurs in unregistering the logger.
     */
    public static Logger unregisterDefaultLogger()
    throws DODSException {
        try {
            Logger logger = defaultLogger;
            turnOffDefaultLogging();
            return logger;
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }

    /**
     * In threading mode, this method unregisters <code>Logger</code>
     * associated with the current thread.
     * In non-threading mode, only default <code>Logger</code>
     * will be unregistered.
     * Call this method to release <code>Logger</code>.
     *
     * @return unregistered <code>Logger</code>.
     *
     * @exception DODSException
     *   If an error occurs in unregistering the logger.
     */
    public static Logger unregisterLogger()
    throws DODSException {
        try {
            Logger logger;
            if (threading)
                return (Logger)loggers.remove(Thread.currentThread());
            logger = defaultLogger;
            turnOffDefaultLogging();
            return logger;
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }

    /**
     * Unregisters <code>Logger</code> associated with the given thread.
     * Call this method to release <code>Logger</code>.
     *
     * @return unregistered <code>Logger</code>.
     *
     * @exception DODSException
     *   If an error occurs in unregistering the logger.
     */
    public static Logger unregisterLogger(Thread thread)
    throws DODSException {
        try {
            return (Logger)loggers.remove(thread);
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }


    /**
     * Returns the default DatabaseManager.
     *
     * @return the default DatabaseManager.
     */
    public static DatabaseManager getDefaultDatabaseManager() {
        return defaultDatabaseManager;
    }

    /**
     * Returns the DatabaseManager object for the current thread.
     * Returns default DatabaseManager if there is no database
     * manager associated with the current thread. Returns null
     * if default DatabaseManager is not set.
     * If non-threading mode is used, default <code>DatabaseManager</code>
     * will be returned.
     *
     * @return the DatabaseManager object, if available, otherwise null.
     */
    public static DatabaseManager getDatabaseManager() {
        if (threading) {
            DatabaseManager dbManager = (DatabaseManager)databaseManagers.get(Thread.currentThread());
            if (dbManager != null)
                return dbManager;
        }
        return defaultDatabaseManager;
    }

    /**
     * Returns the DatabaseManager object for the given thread.
     * Returns default DatabaseManager if there is no database
     * manager associated with the thread. Returns null if default
     * DatabaseManager is not set.
     *
     * @param thread the thread to associate with the <code>DatabaseManager</code>.
     *
     * @return the DatabaseManager object, if available, otherwise null.
     */
    public static DatabaseManager getDatabaseManager(Thread thread) {
        DatabaseManager dbManager = (DatabaseManager)databaseManagers.get(thread);
        if (dbManager != null)
            return dbManager;
        return defaultDatabaseManager;
    }

    /**
     * Returns the default logger.
     *
     * @return The default logger.
     */
    public static Logger getDefaultLogger() {
        return defaultLogger;
    }

    /**
     * Returns the logger object for the current thread.
     * Returns default logger if there is no logger associated with the current
     *  thread. Returns null if default logger is not set.
     * If non-threading mode is used, default <code>Logger</code> will be returned.
     *
     * @return the logger object, if available, otherwise null.
     */
    public static Logger getLogger() {
        if (threading) {
            Logger logger = (Logger)loggers.get(Thread.currentThread());
            if (logger != null)
                return logger;
        }
        return defaultLogger;
    }

    /**
     * Returns the logger object for the given thread.
     * Returns default logger if there is no logger associated with the thread.
     * Returns null if default logger is not set.
     *
     * @param thread the thread to associate with the logger.
     *
     * @return the logger object, if available, otherwise null.
     */
    public static Logger getLogger(Thread thread) {
        Logger logger = (Logger)loggers.get(thread);
        if (logger != null)
            return logger;
        return defaultLogger;
    }

    /**
     * Shutdowns all <code>DatabaseManager</code>s and <code>Loggers</code>s.
     * Call this method to release and shutdown all <code>DatabaseManager</code>s
     * and <code>Loggers</code>s.
     *
     * @exception DODSException
     *   If an error occurs in releasing DatabaseManagers and loggers.
     */
    public static void shutdown()
    throws DODSException {
        try {
            if (threading) {
                Set keys = databaseManagers.keySet();
                for (Iterator iter = keys.iterator();iter.hasNext(); ) {
                    Thread key = (Thread)iter.next();
                    DatabaseManager dbManager = (DatabaseManager)databaseManagers.remove(key);
                    if (dbManager != null) {
                        dbManager.shutdown();
                        dbManager = null;
                    }
                }
                keys = loggers.keySet();
                for (Iterator iter = keys.iterator();iter.hasNext(); ) {
                    Thread key = (Thread)iter.next();
                    loggers.remove(key);
                }
            }
            if (defaultDatabaseManager != null) {
                defaultDatabaseManager.shutdown();
                defaultDatabaseManager = null;
            }
            turnOffDefaultLogging();
        } catch (Exception e) {
            throw new DODSException(e);
        }
    }

    /**
     * Returns the mode of usage. If this flag is <code>true</code>, DODS uses multi
     * threading mode. This means that DODS handles multiple threads, like
     * Enhydra application server (each application in one <code>Thread</code>).
     * If this flag is <code>false</code>, there can be more then one <code>Thread</code>,
     * but there is only one <code>DatabaseManager</code> for all <code>Thread</code>s
     * (for the whole application).
     *
     * @return <code>true</code> if threading mode is used, otherwise
     *         <code>false</code>.
     */
    public static boolean isThreading() {
        return threading;
    }

    /**
     * Sets mode of usage. If this flag is <code>true</code>, DODS uses multi
     * threading mode. This means that DODS handles multiple threads, like
     * Enhydra application server (each application in one <code>Thread</code>).
     * If this flag is <code>false</code>, there can be more then one <code>Thread</code>,
     * but there is only one <code>DatabaseManager</code> for all <code>Thread</code>s
     * (for the whole application).
     *
     * @param mode mode of usage.
     *
     */
    public static void setThreading(boolean mode) {
        threading = mode;
    }

    /**
     * Initializes DODS.
     * Sets default logger.
     *
     */
    protected static void init() {
        threading = false;
        databaseManagers = new HashMap();
        loggers = new HashMap();
        turnOffDefaultLogging();
    }

    /**
     * Sets default Logger. If root Logger does not exist, turns off default
     * Logging.
     */
    static void turnOffDefaultLogging() {
        defaultLogger = Logger.getRootLogger();
        boolean rootIsConfigured = defaultLogger.getAllAppenders().hasMoreElements();
        if(!rootIsConfigured) {
           defaultLogger.setLevel(Level.OFF);
        }
    }
    public static void main (String[] args) {
        try {
            String fileName = "discRack.conf";
            DatabaseManager dbManager = StandardDatabaseManager.newInstance(fileName);
            DODS.register(dbManager);
        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
