/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 */

/*
 *
 * @author    Tanja Jovanovic
 * @version   2.0  07.01.2003.
 *
 */
package org.enhydra.dods.cache;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import com.lutris.dods.builder.generator.dataobject.GenericDO;
import com.lutris.util.Config;

/**
 *
 * This class contains data and mechanisms needed for caching data objects and queries.
 *
 */
public class QueryCacheImpl implements QueryCache {

    /**
     * LRU cache for storing data objects.
     * key: cache handle
     * value: data object
     */
    protected LRUCache cache;

    /**
     * Cache for storing simple queries.
     * key: query in String form
     * value: org.enhydra.dods.cache.QueryItemCache
     */
    protected LRUCache simpleQCache;

    /**
     * Cache for storing complex queries.
     * key: query in String form
     * value: org.enhydra.dods.cache.QueryItemCache
     * (value of conds in QueryItemCache is null for complex queries)
     */
    protected LRUCache complexQCache;

    /**
     * This attribute contains information if cache is read-only.
     * (true - is read-only, otherwise false)
     */
    protected boolean cacheReadOnly = false;

    /**
     * Initial query statement, used for cache initialization.
     */
    protected String initialQueryCache = null;

    /**
     * This attribute contains information if the type of caching is "full".
     */
    protected boolean fullCaching = false;

    /**
     * This attribute contains information if multi databases are used.
     */
    protected boolean multi = false;

    /**
     * Constructor (int) - creates cache with maximal size maxCSize and simple
     * and complex query caches with their default maximal sizes.
     *
     * @param  maxCSize maximal cache size.
     */
    public QueryCacheImpl(int maxCSize) {
        cache = new LRUCache(maxCSize);
        simpleQCache = new LRUCache(DEFAULT_MAX_SIMPLE_QUERY_CACHE_SIZE);
        complexQCache = new LRUCache(DEFAULT_MAX_COMPLEX_QUERY_CACHE_SIZE);
    }

    /**
      * Constructor () - creates cache with its default maximal size and simple
      * and complex query caches with their default maximal sizes.
      *
      * @param  maxCSize maximal cache size.
      */
    public QueryCacheImpl() {
        cache = new LRUCache(DEFAULT_MAX_CACHE_SIZE);
        simpleQCache = new LRUCache(DEFAULT_MAX_SIMPLE_QUERY_CACHE_SIZE);
        complexQCache = new LRUCache(DEFAULT_MAX_COMPLEX_QUERY_CACHE_SIZE);
    }

    /**
     * Refreshes caches.
     *
     * @param  maxCSize maximal cache size.
     */
   public void refresh() {
       cache = new LRUCache(cache.getMaxEntries());
       if (simpleQCache != null)
           simpleQCache = new LRUCache(simpleQCache.getMaxEntries());
       if (complexQCache != null)
           complexQCache = new LRUCache(complexQCache.getMaxEntries());
       checkFull();
   }

    /**
     * Returns cache content.
     *
     * @return cache content as <code>HashMap</code> of DO objects.
     */
    public Map getCacheContent() {
        return cache;
    }

    /**
     * Returns maximum cache size.
     *
     * @return Maximum cache size.
     */
    public int getMaxCacheSize() {
        if (cache != null)
            return cache.getMaxEntries();
        return 0;
    }

    /**
     * Get maximum cache size. If the cache is unbounded (the maximum size
     * is negative), the current cache size is returned.
     *
     * @param real If this parameter is true method return real maximum cache
     * size, othervise return appropriate value for statistic.
     *
     * @return maximum cache size.
     */
    public int getMaxCacheSize(boolean real) {
        int size = getMaxCacheSize();
        if (size < 0) {
            if (real)
                return -1;
            else
                return getCacheSize();
        }
        return size;
    }

    /**
     * Returns size of currently used cache.
     *
     * @return Size of currently used cache.
     */
    public int getCacheSize() {
        if (cache != null)
            return cache.size();
        return 0;
    }

    /**
     * Sets maximum cache size.
     *
     * @param maxSize Maximum cache size.
     */
    public synchronized void setMaxCacheSize(int maxSize) {
        if (maxSize == 0) {
            cache = null;
            if (toReconfigure())
        	    checkFull();
            return;
        }
        if (cache == null)
            cache = new LRUCache(maxSize);
        else
            cache.setMaxEntries(maxSize);
        if (toReconfigure())
        	  checkFull();
    }

    /**
     * Returns maximum size of simple query cache.
     *
     * @return Maximum size of simple query cache.
     */
    public int getMaxSimpleQueryCacheSize() {
        if (simpleQCache != null)
            return simpleQCache.getMaxEntries();
        return 0;
    }

    /**
     * Gets maximum simple query cache size. If the cache is unbounded
     * (the maximum size is negative), the current cache size is returned.
     *
     * @param real If this parameter is true method return real maximum cache
     * size, othervise return appropriate value for statistic.
     *
     * @return maximum simple query cache size.
     */
    public int getMaxSimpleQueryCacheSize(boolean real) {
        int size = getMaxSimpleQueryCacheSize();
        if (size < 0) {
            if (real)
                return -1;
            else
                return getSimpleQueryCacheSize();
        }
        return size;
    }

    /**
     * Sets maximum size of simple query cache.
     *
     * @param maxSize maximum size of simple query cache.
     */
    public synchronized void setMaxSimpleQueryCacheSize(int maxSize) {
        if (maxSize == 0) {

            simpleQCache = null;
            return;
        }
        if (simpleQCache == null)
            simpleQCache = new LRUCache(maxSize);
        else
            simpleQCache.setMaxEntries(maxSize);
    }

    /**
     * Returns size of currently used simple query cache.
     *
     * @return Size of currently used simple query cache.
     */
    public int getSimpleQueryCacheSize() {
        if (simpleQCache != null)
            return simpleQCache.size();
        return 0;
    }

    /**
     * Returns maximum size of complex query cache.
     *
     * @return Maximum size of complex query cache.
     */
    public int getMaxComplexQueryCacheSize() {
        if (complexQCache != null)
            return complexQCache.getMaxEntries();
        return 0;
    }

    /**
     * Gets maximum complex query cache size. If the cache is unbounded
     * (the maximum size is negative), the current cache size is returned.
     *
     * @param real If this parameter is true method return real maximum cache
     * size, othervise return appropriate value for statistic.
     *
     * @return maximum complex query cache size.
     */
    public int getMaxComplexQueryCacheSize(boolean real) {
        int size = getMaxComplexQueryCacheSize();
        if (size < 0) {
            if (real)
                return -1;
            else
                return getComplexQueryCacheSize();
        }
        return size;
    }


    /**
     * Sets maximum size of complex query cache.
     *
     * @param maxSize maximum size of complex query cache.
     */
    public synchronized void setMaxComplexQueryCacheSize(int maxSize) {
        if (maxSize == 0) {
            complexQCache = null;
            return;
        }
        if (complexQCache == null)
            complexQCache = new LRUCache(maxSize);
        else
            complexQCache.setMaxEntries(maxSize);
    }

    /**
     * Returns size of currently used complex query cache.
     *
     * @return Size of currently used complex query cache.
     */
    public int getComplexQueryCacheSize() {
        if (complexQCache != null)
            return complexQCache.size();
        return 0;
    }

    /**
     * Sets read-only attribute.
     *
     * @param readOnly New read-only value.
     */
    public synchronized void setCacheReadOnly(boolean readOnly) {
        cacheReadOnly = readOnly;
        if (toReconfigure())
        	  checkFull();
    }

    /**
     * Returns read-only attribute.
     *
     * @return read-only attribute
     */
    public boolean getCacheReadOnly() {
        return cacheReadOnly;
    }

    /**
     * Sets initialQueryCache attribute.
     *
     * @param initQ New initialQueryCache.
     */
    public synchronized void setInitialQueryCache(String initQ) {
        initialQueryCache = initQ;
        if (toReconfigure())
        	  checkFull();
    }

    /**
    * Returns initialQueryCache.
    *
    * @return initialQueryCache.
    */
    public String getInitialQueryCache() {
        return initialQueryCache;
    }

    /**
     * Returns
     * @return
     */
    public boolean isFull() {
        return fullCaching;
    }

    /**
     * @return true if multi caching is used, otherwise false.
     */
    public boolean isMulti() {
        return multi;
    }

    /**
     * @param newMulti new value for multi
     */
    public void setMulti(boolean newMulti) {
        multi = newMulti;
    }

    /**
     * Returns cache.
     * @return Cache.
     */
    public LRUCache getCache() {
        return cache;
    }

    /**
     * Checks wheather cache reconfiguration needs to be done.
     *
     * @return true if cache reconfiguration needs to be done, otherwise false.
     */
    public boolean toReconfigure() {
        return false;
    }

    /**
     * Read cache configuration from application's configuration file.
     *
     * @param config configuration for this cache.
     * @param globalConfig default configuration for all application's caches.
     */
    public void readConfiguration(Config config, Config globalConfig) {
        int maxSimple = QueryCache.DEFAULT_MAX_SIMPLE_QUERY_CACHE_SIZE;
        int maxComplex = QueryCache.DEFAULT_MAX_COMPLEX_QUERY_CACHE_SIZE;
        if (globalConfig != null) {
            try {
                maxSimple = globalConfig.getInt("defaultMaxSimpleQueryCacheSize");
            } catch ( Exception e ) {
            }
            try {
               maxComplex = globalConfig.getInt("defaultMaxComplexQueryCacheSize");
           } catch ( Exception e ) {
           }
            try {
               cacheReadOnly = globalConfig.getBoolean("defaultReadOnlyCache");
           } catch ( Exception e ) {
           }
        }
        if (config != null) {
            try {
                maxSimple = config.getInt("maxSimpleQueryCacheSize");
            } catch ( Exception e ) {
            }
            try {
                maxComplex = config.getInt("maxComplexQueryCacheSize");
            } catch ( Exception e ) {
            }
            try {
                cacheReadOnly = config.getBoolean("readOnlyCache");
            } catch ( Exception e ) {
            }
            try {
                initialQueryCache = config.getString("initalConditionForCache");
            } catch ( Exception e ) {
            }
        }
        this.setMaxSimpleQueryCacheSize(maxSimple);
        this.setMaxComplexQueryCacheSize(maxComplex);
        checkFull();
//        show();
    }

    /**
     * Checks if this is special type of caching - full caching and updates
     * attribute fullCaching if necessary.
     */
    protected void checkFull() {
        if ((getMaxCacheSize() < 0) && (cacheReadOnly == true) && (initialQueryCache != null) && (initialQueryCache.equals("*")))
            fullCaching = true;
        else
            fullCaching = false;
    }

    /**
     *
     *
     */
    public synchronized GenericDO addDO(GenericDO newDO) {
        if (cache == null)
            return null;
        try {
            String handle = newDO.getCacheHandle();
            GenericDO ret = (GenericDO)cache.add(handle, newDO);
            if (ret != null)
                if (ret.getCacheHandle().equals(handle)) {
                    ret.setData(newDO.getData()); // change old references
                    return ret;
                }
            return newDO;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Remove DO from cache.
     *
     * @param cacheHandle Key of the object that will be removed.
     *
     * @return removed DO object.
     */
    public synchronized GenericDO removeDO(GenericDO DO) {
        if (cache != null) {
            try {
                String handle = DO.getCacheHandle();
                return (GenericDO)cache.remove(handle);

            } catch (Exception e) {
            }
        }
        return null;
    }

    /**
     * Remove DO from cache.
     *
     * @param handle String representation of object Id that will be removed.
     *
     * @return removed DO object.
     */
    public synchronized GenericDO removeDO(String handle) {
        if (cache != null) {
            try {
                return (GenericDO)cache.remove(handle);
            } catch (Exception e) {
            }
        }
        return null;
    }


    /**
     * Get cache type.
     *
     * @return Cache type.
     */
    public String getCacheType() {
        if (getMaxCacheSize() == 0)
            return "none";
        else {
            if (fullCaching)
                return "full";
            else
                return "lru";
        }
    }

    /**
     *
     */
    public GenericDO updateDO(GenericDO DO) {
        DO = addDO(DO);
        QueryCacheItem queryItem = null;
        HashSet del;
        if (complexQCache != null) {
            del = new HashSet();
            for (Iterator iter = complexQCache.values().iterator(); iter.hasNext(); ) {
                queryItem = (QueryCacheItem)iter.next();
                if (DO.getOriginDatabase().equals(queryItem.getOriginDatabase()))
                    del.add(queryItem);
            }
            for (Iterator iter = del.iterator(); iter.hasNext(); )
                removeComplexQuery((QueryCacheItem)iter.next());
        }
        if (simpleQCache != null) {
            Iterator iter = null;
            for (iter = simpleQCache.values().iterator(); iter.hasNext();) {
                queryItem = (QueryCacheItem)iter.next();
                queryItem.update(DO);
            }
        }
        return DO;
    }

    /**
     * Deletes DO from simpleQCache and complexQCache
     */
    public GenericDO deleteDO(GenericDO DO) {
        GenericDO oldDO = removeDO(DO);
        Iterator iter = null;
        QueryCacheItem queryItem = null;
        if (simpleQCache != null) {
            for (iter = simpleQCache.values().iterator(); iter.hasNext();) {
                queryItem = (QueryCacheItem)iter.next();
                queryItem.delete(DO);
            }
        }
        if (complexQCache != null) {
            for (iter = complexQCache.values().iterator(); iter.hasNext();) {
                queryItem = (QueryCacheItem)iter.next();
                queryItem.delete(DO);
            }
        }
        return oldDO;
    }


    public QueryCacheItem newQueryCacheItemInstance(String dbName) {
        return new QueryCacheItemImpl(dbName);
    }

    public synchronized QueryCacheItem addSimpleQuery(QueryCacheItem queryItem) {
        if (simpleQCache != null)
            return (QueryCacheItem)simpleQCache.add(
                    queryItem.getOriginDatabase() + "." + queryItem.getQueryId(),
                    queryItem);
        return null;
    }

    public synchronized QueryCacheItem removeSimpleQuery(QueryCacheItem queryItem) {
        if (simpleQCache != null)
            return (QueryCacheItem)simpleQCache.remove(
                    queryItem.getOriginDatabase() +  "." + queryItem.getQueryId());
        return null;
    }

    public synchronized QueryCacheItem addComplexQuery(QueryCacheItem queryItem) {
        if (complexQCache != null)
            return (QueryCacheItem)complexQCache.add(
                    queryItem.getOriginDatabase() +"." + queryItem.getQueryId(),
                    queryItem);
        return null;
    }

    public synchronized QueryCacheItem removeComplexQuery(QueryCacheItem queryItem) {
        if (complexQCache != null)
            return (QueryCacheItem)complexQCache.remove(
                    queryItem.getOriginDatabase() + "." + queryItem.getQueryId());
        return null;
    }

    /**
     * @param handle
     * @return
     */
    public synchronized GenericDO getDOByHandle(String handle) {
        if (cache == null)
            return null;
        return (GenericDO)cache.get(handle);
    }

    public synchronized QueryResult getSimpleQueryResults(String dbName, String query) {
        if (simpleQCache == null)
            return null;
        QueryResult result = null;
        String queryHandle = dbName + "." + query;
        QueryCacheItem queryItem = (QueryCacheItem)simpleQCache.get(queryHandle);
        if (queryItem != null) {
            result = new QueryResult();
            result.database = queryItem.getOriginDatabase();
            DOShell shell = null;
            String handle = null;
            String cachePrefix = "";
            if (isMulti())
                cachePrefix = queryItem.getOriginDatabase() + ".";
            GenericDO cacheDO = null;
            for (Iterator iter = queryItem.getOIds().iterator(); iter.hasNext(); ) {
                handle = (String)iter.next();
                try {
                    cacheDO = (GenericDO)cache.get(cachePrefix+handle);
                    shell = new DOShell(handle);
                    if (cacheDO != null)
                        shell.dataObject = cacheDO;
                    else
                        result.lazy.add(shell);
                    result.DOs.add(shell);
                } catch (Exception e) {
                }
            }
        }
        return result;
    }

    public synchronized QueryResult getComplexQueryResults(String dbName, String query) {
        if (complexQCache == null)
           return null;
       QueryResult result = null;
       String queryHandle = dbName + "." + query;
       QueryCacheItem queryItem = (QueryCacheItem)complexQCache.get(queryHandle);
       if (queryItem != null) {
           result = new QueryResult();
           result.database = queryItem.getOriginDatabase();
           DOShell shell = null;
           String handle = null;
           String cachePrefix = "";
           if (isMulti())
               cachePrefix = queryItem.getOriginDatabase() + ".";
           GenericDO cacheDO = null;
           for (Iterator iter = queryItem.getOIds().iterator(); iter.hasNext(); ) {
               handle = (String)iter.next();
               try {
                   cacheDO = (GenericDO)cache.get(cachePrefix+handle);
                   shell = new DOShell(handle);
                   if (cacheDO != null)
                       shell.dataObject = cacheDO;
                   else
                       result.lazy.add(shell);
                   result.DOs.add(shell);
               } catch (Exception e) {
               }
           }
       }
       return result;
    }

    public synchronized QueryResult getQueryResults(String dbName, String query) {
        QueryResult result = getSimpleQueryResults(dbName, query);
        if (result == null)
            result = getComplexQueryResults(dbName, query);
        return result;
    }


	/**
	 * Shows content of this class.
	 * Can be used for debugging.
	 */
	public void show() {
		System.out.println("-------------------------------------------------");
//		System.out.println(" cache: " + cache);
//		System.out.println(" simpleQCache : " + simpleQCache);
//		System.out.println(" complexQCache : " + complexQCache);
		System.out.println(" maxCacheSize : " + getMaxCacheSize());
		System.out.println(" cacheReadOnly : " + cacheReadOnly);
		System.out.println(" initialQueryCache : " + initialQueryCache);
		System.out.println(" fullCaching : " + fullCaching);
 	}

	/**
	 * Shows content of this class.
	 * Can be used for debugging.
	 */
	public String toString() {
		StringBuffer ret = new StringBuffer();
		ret.append("\n QueryCacheImpl: ");
		ret.append("\n cache: " + cache);
		ret.append("\n simpleQCache : " + simpleQCache);
		ret.append("\n complexQCache : " + complexQCache);
		ret.append("\n cacheReadOnly : " + cacheReadOnly);
		ret.append("\n initialQueryCache : " + initialQueryCache);
		ret.append("\n fullCaching : " + fullCaching);
        return ret.toString();
 	}


}
