
/**
 * Title:        Generic Datastore<p>
 * Description:  Extension for all DODS DO's to store variable attributes which are not defined in a doml file<p>
 * Copyright:    Copyright (c) Together Teamlsungen EDV-Dienstleistungen GmbH. under LGPL<p>
 * Company:      Together Teamlsungen EDV-Dienstleistungen GmbH.<p>
 * @author Alfred Madl (a.madl@together.at)
 * @version 1.0
 */
package org.webdocwf.util.genericstore;

import java.math.BigDecimal;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import org.enhydra.dods.DODS;
import org.webdocwf.dods.access.AccessEvalException;
import org.webdocwf.dods.access.AccessException;
import org.webdocwf.dods.access.AccessRightException;
import org.webdocwf.dods.access.SecureDO;
import org.webdocwf.dods.access.User;
import org.webdocwf.util.genericstore.data.AttributeBinaryValueDO;
import org.webdocwf.util.genericstore.data.AttributeBinaryValueQuery;
import org.webdocwf.util.genericstore.data.AttributeBoolValueDO;
import org.webdocwf.util.genericstore.data.AttributeBoolValueQuery;
import org.webdocwf.util.genericstore.data.AttributeDateValueDO;
import org.webdocwf.util.genericstore.data.AttributeDateValueQuery;
import org.webdocwf.util.genericstore.data.AttributeDefinitionDO;
import org.webdocwf.util.genericstore.data.AttributeDefinitionQuery;
import org.webdocwf.util.genericstore.data.AttributeEnumValueDO;
import org.webdocwf.util.genericstore.data.AttributeEnumValueQuery;
import org.webdocwf.util.genericstore.data.AttributeNumberValueDO;
import org.webdocwf.util.genericstore.data.AttributeNumberValueQuery;
import org.webdocwf.util.genericstore.data.AttributeObjectValueDO;
import org.webdocwf.util.genericstore.data.AttributeObjectValueQuery;
import org.webdocwf.util.genericstore.data.AttributeStringValueDO;
import org.webdocwf.util.genericstore.data.AttributeStringValueQuery;
import org.webdocwf.util.genericstore.data.EnumTypeDefinitionDO;
import org.webdocwf.util.genericstore.data.EnumTypeDefinitionQuery;
import org.webdocwf.util.genericstore.data.EnumValueDefinitionDO;
import org.webdocwf.util.genericstore.data.EnumValueDefinitionQuery;
import org.webdocwf.util.genericstore.data.GenericObjectDO;
import org.webdocwf.util.genericstore.data.GenericObjectQuery;
import org.webdocwf.util.genericstore.data.StateTransitionDO;
import org.webdocwf.util.genericstore.data.StateTransitionQuery;
import org.webdocwf.util.genericstore.data.TypeDefinitionDO;
import org.webdocwf.util.genericstore.data.TypeDefinitionQuery;

import com.lutris.appserver.server.sql.DBRowUpdateException;
import com.lutris.appserver.server.sql.DBTransaction;
import com.lutris.appserver.server.sql.DatabaseManagerException;
import com.lutris.appserver.server.sql.ObjectId;
import com.lutris.appserver.server.sql.ObjectIdException;
import com.lutris.dods.builder.generator.query.DataObjectException;
import com.lutris.dods.builder.generator.query.NonUniqueQueryException;
import com.lutris.dods.builder.generator.query.QueryException;
import com.lutris.dods.builder.generator.query.RefAssertionException;

public abstract class GenericStoreDO extends SecureDO {

  /**
   * Abstract method which must be implemented by subclasses (templates)
   * and call the static createExisting(bd) method !
   *
   * @param bd ObjectID as BigDecimal of the object which must be read from the DB
   * @return The object which was read from the DB
   */
  public abstract GenericStoreDO createForExisting( BigDecimal bd )
    throws SQLException, ObjectIdException, DataObjectException, DatabaseManagerException;

  /**
   * Abstract method which must be implemented by subclasses (templates)
   * Is necessary for SecureDO implementation
   *
   * @param usr The user which is asked for the access right
   * @return Whether the user is allowed to read the DO existance
   */
  public GenericStoreDO() throws DatabaseManagerException, ObjectIdException {
  	super();
  }
  public GenericStoreDO(String dbName) throws DatabaseManagerException, ObjectIdException {
	super(dbName);
  }

  public GenericStoreDO( boolean isView ) throws ObjectIdException, DatabaseManagerException {
        super(isView);
  }
  public GenericStoreDO(String dbName, boolean isView ) throws ObjectIdException, DatabaseManagerException {
        super(dbName,isView);
  }

  public GenericStoreDO( ObjectId id ) throws ObjectIdException, DatabaseManagerException {
        super(id);
  }
  public GenericStoreDO( String dbName, ObjectId id ) throws ObjectIdException, DatabaseManagerException {
        super(dbName, id);
  }

  public GenericStoreDO(ResultSet rs) throws SQLException, ObjectIdException, DatabaseManagerException {
        super(rs);
  }
  public GenericStoreDO(String dbName, ResultSet rs) throws SQLException, ObjectIdException, DatabaseManagerException {
        super(dbName,rs);
  }

  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetEnumAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetEnumAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericEnum", null);
   }
  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetDateAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetDateAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericDate", null);
   }
  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetBoolAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetBoolAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericBool", null);
   }
  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetNumberAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetNumberAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericNumber", null);
   }
  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetStringAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetStringAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericString", null);
   }
  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetBinaryAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetBinaryAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericBinary", null);
   }
  /**
   * Ensure that the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetObjectAttrAccess( String attrKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericGetObjectAttrAccess( attrKey, usr )) throw new AccessRightException("No access !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericPointer", null);
   }
  /**
   * Ensure that the user is allowed to read the existence of the object
   * which is referenced by a pointer
   *
   * @param attrKey The name of the attribute (pointer) which will be read
   * @param value The DO which will be read
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericGetDOValueAccess( String attrKey, SecureDO value, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	checkAttrDefinition(attrKey);
	if (value != null)
	{
	  if (!value.hasDOGetAccess( usr )) throw new AccessRightException("No access to object in pointer !", usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericPointer", null);
	}
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetEnumAttrAccess( String attrKey, String newValueKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetEnumAttrAccess( attrKey, null, newValueKey, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValueKey, null, "GenericDate", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value and type
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param newType The new attribute type wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetEnumAttrAccess( String attrKey, String newValueType, String newValueKey, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetEnumAttrAccess( attrKey, newValueType, newValueKey, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValueKey, null, "GenericDate", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetDateAttrAccess( String attrKey, java.sql.Timestamp newValue, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetDateAttrAccess( attrKey, newValue, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValue.toString(), null, "GenericDate", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetBoolAttrAccess( String attrKey, Boolean newValue, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetBoolAttrAccess( attrKey, newValue, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValue.toString(), null, "GenericBool", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetNumberAttrAccess( String attrKey, java.math.BigDecimal newValue, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetNumberAttrAccess( attrKey, newValue, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValue.toString(), null, "GenericNumber", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetStringAttrAccess( String attrKey, String newValue, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetStringAttrAccess( attrKey, newValue, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValue, null, "GenericString", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetBinaryAttrAccess( String attrKey, byte[] newValue, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetBinaryAttrAccess( attrKey, newValue, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValue.toString(), null, "GenericBinary", null);
   }
  /**
   * Ensure that the user is allowed to set the attribute
   * with a certain value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new attribute value wich will be set
   * @param usr The user which is asked for the access right
   */
   public void assertDOGenericSetObjectAttrAccess( String attrKey, GenericStoreDO newValue, org.webdocwf.dods.access.User usr )
   throws AccessException, GenericStoreDefinitionException, DataObjectException {
	if (!hasDOGenericSetObjectAttrAccess( attrKey, newValue, usr )) throw new AccessRightException("No access !", usr, "GenericSetAttr", this, null, attrKey, newValue.getClass().getName()+" : " +newValue.getOId().toString(), null, "GenericPointer", null);
   }
  /**
   * Read a certain generic attribute enum type
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute enum type of the DO
   */
   public String genericGetEnumTypeAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetEnumAttrAccess( attrKey, usr );
      String ret = genericGetEnumTypeAttr(attrKey);
      return ret;
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public String genericGetEnumAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetEnumAttrAccess( attrKey, usr );
      String ret = genericGetEnumAttr(attrKey);
      return ret;
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public java.sql.Timestamp genericGetDateAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetDateAttrAccess( attrKey, usr );
      java.sql.Timestamp ret = genericGetDateAttr(attrKey);
      return ret;
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public Boolean genericGetBoolAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetBoolAttrAccess( attrKey, usr );
      Boolean ret = genericGetBoolAttr(attrKey);
      return ret;
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public String genericGetStringAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetStringAttrAccess( attrKey, usr );
      String ret = genericGetStringAttr(attrKey);
      return ret;
   }

  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public byte[] genericGetBinaryAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetBinaryAttrAccess( attrKey, usr );
      byte[] ret = genericGetBinaryAttr(attrKey);
      return ret;
   }

  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public java.math.BigDecimal genericGetNumberAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetNumberAttrAccess( attrKey, usr );
      java.math.BigDecimal ret = genericGetNumberAttr(attrKey);
      return ret;
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return The desired generic attribute value of the DO
   */
   public GenericStoreDO genericGetObjectAttr ( String attrKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericGetObjectAttrAccess( attrKey, usr );
      GenericStoreDO ret = genericGetObjectAttr(attrKey);
      assertDOGenericGetDOValueAccess( attrKey, ret, usr );
      return ret;
   }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetEnumAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      String value;
      String type;
      try {
        value = genericGetEnumAttr(attrKey);
        type = genericGetEnumTypeAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericEnum", null);
      }
      return hasDOGenericGetAttrAccess( attrKey, type, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetDateAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      java.sql.Timestamp value;
      try {
        value = genericGetDateAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericDate", null);
      }
      return hasDOGenericGetAttrAccess( attrKey, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetBoolAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      Boolean value;
      try {
        value = genericGetBoolAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericBool", null);
      }
      return hasDOGenericGetAttrAccess( attrKey, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetNumberAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
	java.math.BigDecimal value;
      try {
        value = genericGetNumberAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericNumber", null);
 	 }
	 return hasDOGenericGetAttrAccess( attrKey, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetStringAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
	String value;
      try {
        value = genericGetStringAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericString", null);
 	 }
	 return hasDOGenericGetAttrAccess( attrKey, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetBinaryAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
	byte[] value;
      try {
        value = genericGetBinaryAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericBinary", null);
 	 }
	 return hasDOGenericGetAttrAccess( attrKey, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   *
   * @param attrKey The name of the attribute which will be modified
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute
   */
    public boolean hasDOGenericGetObjectAttrAccess( String attrKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
	GenericStoreDO value;
      try {
        value = genericGetObjectAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericGetAttr", this, null, attrKey, null, null, "GenericPointer", null);
 	 }
	 return hasDOGenericGetAttrAccess( attrKey, value, usr );
    }
  /**
   * Check whether the user is allowed to read the attribute
   * with a certain current value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param value The current value of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute value
   */
    public boolean hasDOGenericGetAttrAccess( String attrKey, Object value, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
	checkAttrDefinition(attrKey);
	return usr.hasDOGetAttrAccess ( this, attrKey, value );
    }
  /**
   * Check whether the user is allowed to read the attribute
   * with a certain current value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param value The current value of the attribute which will be read
   * @param type The current type of the attribute which will be read
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to read the attribute value
   */
    public boolean hasDOGenericGetAttrAccess( String attrKey, String type, String value, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
	checkAttrDefinition(attrKey);
	if (value!=null) checkEnumValueDefinition(type, value);
	return usr.hasDOGetAttrAccess ( this, attrKey, type, value );
    }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetEnumAttrAccess( String attrKey, String newValueType, String newValueKey, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      String value;
      String type;
      try {
        value = genericGetEnumAttr(attrKey);
        type = genericGetEnumTypeAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericEnum", null);
      }
      if (newValueType == null) newValueType = type;
	return hasDOGenericSetAttrAccess( attrKey, type, value, newValueType, newValueKey, usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetDateAttrAccess( String attrKey, java.sql.Timestamp newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      java.sql.Timestamp value;
      try {
        value = genericGetDateAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericDate", null);
      }
	return hasDOGenericSetAttrAccess( attrKey, value,newValue,usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetBoolAttrAccess( String attrKey, Boolean newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      Boolean value;
      try {
        value = genericGetBoolAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericBool", null);
      }
	return hasDOGenericSetAttrAccess( attrKey, value,newValue,usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetNumberAttrAccess( String attrKey, java.math.BigDecimal newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      java.math.BigDecimal value;
      try {
        value = genericGetNumberAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericNumber", null);
      }
	return hasDOGenericSetAttrAccess( attrKey, value,newValue,usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetStringAttrAccess( String attrKey, String newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      String value;
      try {
        value = genericGetStringAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericString", null);
      }
	return hasDOGenericSetAttrAccess( attrKey, value,newValue,usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetBinaryAttrAccess( String attrKey, byte[] newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      byte[] value;
      try {
        value = genericGetBinaryAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericBinary", null);
      }
	return hasDOGenericSetAttrAccess( attrKey, value,newValue,usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
   public boolean hasDOGenericSetObjectAttrAccess( String attrKey, GenericStoreDO newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException {
      GenericStoreDO value;
      try {
        value = genericGetObjectAttr(attrKey);
      } catch (Exception e) {
		throw new AccessEvalException("Error in access evaluation !", e, usr, "GenericSetAttr", this, null, attrKey, null, null, "GenericPointer", null);
      }
	return hasDOGenericSetAttrAccess( attrKey, value,newValue,usr );
   }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value and a certain current value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param oldValue The current (old) value wich will be overwritten
   * @param oldType The current (old) type wich will be overwritten
   * @param newValue The new value wich will be written to the DO
   * @param newType The new type wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
    public boolean hasDOGenericSetAttrAccess( String attrKey, String oldType, String oldValue, String newType, String newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException
    {
	checkAttrDefinition(attrKey);
	if (oldValue!=null) checkEnumValueDefinition(oldType, oldValue);
	if (newValue!=null) checkEnumValueDefinition(newType, newValue);
	return usr.hasDOSetAttrAccess ( this, attrKey, oldType, oldValue, newType, newValue);
    }
  /**
   * Check whether the user is allowed to set the attribute
   * with a certain new value and a certain current value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param oldValue The current (old) value wich will be overwritten
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   * @return Whether the user has the right to set the attribute
   */
    public boolean hasDOGenericSetAttrAccess( String attrKey, Object oldValue, Object newValue, org.webdocwf.dods.access.User usr)
    throws AccessEvalException, GenericStoreDefinitionException, DataObjectException
    {
	checkAttrDefinition(attrKey);
	return usr.hasDOSetAttrAccess ( this, attrKey, oldValue, newValue );
    }
  /**
   * Set a certain generic state attribute with a state
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetStateAttr ( String attrKey, String newValueKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, StateTransitionException, GenericStoreDefinitionException {
      assertDOGenericSetEnumAttrAccess( attrKey, newValueKey, usr );
      genericSetStateAttr( attrKey, newValueKey );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetEnumAttr ( String attrKey, String newValueKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetEnumAttrAccess( attrKey, newValueKey, usr );
      genericSetEnumAttr( attrKey, newValueKey );
   }
  /**
   * Set a certain generic state attribute with a start-value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param startStateKey The start-value (state) wich will be written to the DO
   * @param startTypeKey The type (state machine name) wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetStartStateAttr ( String attrKey, String startTypeKey, String startStateKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, StateTransitionException, GenericStoreDefinitionException {
      assertDOGenericSetEnumAttrAccess( attrKey, startTypeKey, startStateKey, usr );
      genericSetStartStateAttr( attrKey, startTypeKey, startStateKey );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValueKey The new value wich will be written to the DO
   * @param newtypeKey The new type wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetEnumAttr ( String attrKey, String newTypeKey, String newValueKey, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetEnumAttrAccess( attrKey, newTypeKey, newValueKey, usr );
      genericSetEnumAttr( attrKey, newTypeKey , newValueKey );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetDateAttr ( String attrKey, java.sql.Timestamp newValue, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetDateAttrAccess( attrKey, newValue, usr );
      genericSetDateAttr( attrKey, newValue );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetBoolAttr ( String attrKey, Boolean newValue, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetBoolAttrAccess( attrKey, newValue, usr );
      genericSetBoolAttr( attrKey, newValue );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetNumberAttr ( String attrKey, java.math.BigDecimal newValue, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetNumberAttrAccess( attrKey, newValue, usr );
      genericSetNumberAttr( attrKey, newValue );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetStringAttr ( String attrKey, String newValue, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetStringAttrAccess( attrKey, newValue, usr );
      genericSetStringAttr( attrKey, newValue );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetBinaryAttr ( String attrKey, byte[] newValue, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetBinaryAttrAccess( attrKey, newValue, usr );
      genericSetBinaryAttr( attrKey, newValue );
   }
  /**
   * Set a certain generic attribute with a new value
   *
   * @param attrKey The name of the attribute which will be modified
   * @param newValue The new value wich will be written to the DO
   * @param usr The user which is asked for the access right
   */
   public void genericSetObjectAttr ( String attrKey, GenericStoreDO newValue, org.webdocwf.dods.access.User usr )
   throws DataObjectException, AccessException, GenericStoreDefinitionException {
      assertDOGenericSetObjectAttrAccess( attrKey, newValue, usr );
      genericSetObjectAttr( attrKey, newValue );
   }
  /**
   * Check whether the enum attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericEnumLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataEnum ) loadGenericEnumData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load Enum data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(), e );
    }
  }
  /**
   * Check whether the date attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericDateLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataDate ) loadGenericDateData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load Date data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  /**
   * Check whether the bool attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericBoolLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataBool ) loadGenericBoolData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load Bool data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  /**
   * Check whether the number attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericNumberLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataNumber ) loadGenericNumberData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load Number data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  /**
   * Check whether the string attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericStringLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataString ) loadGenericStringData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load String data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  /**
   * Check whether the binary attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericBinaryLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataBinary ) loadGenericBinaryData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load Binary data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  /**
   * Check whether the pointer attributes are loaded
   * if not, load them
   *
   */
  private void checkGenericObjectLoad()
  throws DataObjectException {
    try {
      checkGenericObjectDOLoad();
      if ( null == dataObject ) loadGenericObjectData();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load Object data for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  /**
   * Check whether the object extension is loaded
   * if not, load it
   *
   */
  private void checkGenericObjectDOLoad()
  throws DataObjectException {
    try {
      if ( null == objectExtension ) loadGenericObjectDO();
    } catch ( Exception e ) {
	throw new DataObjectException(
	    "Unable to load GenericObjectDO for GenericStoreDO id=" + getOId() +
	    ", error = " + e.getMessage(),e );
    }
  }
  private GenericObjectDO objectExtension=null;
  private TreeMap dataDate=null;
  private TreeMap dataBool=null;
  private TreeMap dataNumber=null;
  private TreeMap dataString=null;
  private TreeMap dataObject=null;
  private TreeMap dataEnum=null;
  private TreeMap typeEnum=null;
  private TreeMap dataBinary=null;
  /**
   * Read a certain generic attribute type
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public String genericGetEnumTypeAttr (String attrKey)
   throws DataObjectException {
      checkGenericEnumLoad();
      return (String) typeEnum.get(attrKey);
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public String genericGetEnumAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
      checkGenericEnumLoad();
      return (String) dataEnum.get(attrKey);
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public java.sql.Timestamp genericGetDateAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
      checkGenericDateLoad();
      return (java.sql.Timestamp) dataDate.get(attrKey);
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public Boolean genericGetBoolAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
      checkGenericBoolLoad();
      return (Boolean) dataBool.get(attrKey);
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public String genericGetStringAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
      checkGenericStringLoad();
      return (String) dataString.get(attrKey);
   }

  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public java.math.BigDecimal genericGetNumberAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
      checkGenericNumberLoad();
      return (java.math.BigDecimal) dataNumber.get(attrKey);
   }
  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public byte[] genericGetBinaryAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
        checkGenericBinaryLoad();
      return (byte[]) dataBinary.get(attrKey);
   }
  /**
   * Set a certain generic state attribute with a start value
   *
   * @param attrKey The name of the attribute which will be read
   * @param startValueKey The value which is written to the DO
   * @param startTypeKey The type which is written to the DO
   */
   public void genericSetStartStateAttr ( String attrKey, String startTypeKey , String startValueKey )
   throws DataObjectException, StateTransitionException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
	checkEnumValueDefinition(startTypeKey, startValueKey);
	if (genericIsBeginState(startTypeKey, startValueKey))
		genericSetEnumAttr( attrKey, startTypeKey , startValueKey );
	else
		throw new StateTransitionException("Illegal start state "+startTypeKey+": "+startValueKey);
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValueKey The value which is written to the DO
   * @param newTypeKey The type which is written to the DO
   */
   public void genericSetEnumAttr ( String attrKey, String newTypeKey , String newValueKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
	if (newValueKey!=null) checkEnumValueDefinition(newTypeKey, newValueKey);
     try {
      checkGenericEnumLoad();
      String oldValue = (String)dataEnum.get(attrKey);
      String oldType = (String)typeEnum.get(attrKey);
      if ( (oldValue!=null||newValueKey!=null) && newTypeKey!=null && (oldValue==null||!oldValue.equals(newValueKey)))
      {
        dataEnum.put(attrKey, newValueKey);
        typeEnum.put(attrKey, newTypeKey);
        markEnumDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetEnumAttr ( String attrKey, String newValueKey )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
	String newTypeKey = genericGetEnumTypeAttr(attrKey);
	if (newTypeKey==null) newTypeKey=findAttrDef(attrKey).getENUMTYPEDEFINITIONOID().getKEYVALUE();
	if (newValueKey!=null) checkEnumValueDefinition(newTypeKey ,newValueKey);
     try {
      checkGenericEnumLoad();
      String oldValue = (String)dataEnum.get(attrKey);
      if ( (oldValue!=null||newValueKey!=null) && (oldValue==null||!oldValue.equals(newValueKey)))
      {
        dataEnum.put(attrKey, newValueKey);
        typeEnum.put(attrKey, newTypeKey);
        markEnumDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic state attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetStateAttr ( String attrKey, String newValueKey )
   throws DataObjectException, StateTransitionException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
	checkEnumValueDefinition(genericGetEnumTypeAttr(attrKey),newValueKey);
	if (genericGetAllowedTargetStates(attrKey).contains(newValueKey))
		genericSetEnumAttr( attrKey, newValueKey);
	else
		throw new StateTransitionException("Illegal state transition to "+newValueKey);
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetDateAttr ( String attrKey, java.sql.Timestamp newValue )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
     try {
      checkGenericDateLoad();
      java.sql.Timestamp oldValue = (java.sql.Timestamp)dataDate.get(attrKey);
      if ( (oldValue!=null||newValue!=null) && (oldValue==null||!oldValue.equals(newValue)))
      {
        dataDate.put(attrKey, newValue);
        markDateDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetBoolAttr ( String attrKey, Boolean newValue )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
     try {
      checkGenericBoolLoad();
      Boolean oldValue = (Boolean)dataBool.get(attrKey);
      if ( (oldValue!=null||newValue!=null) && (oldValue==null||!oldValue.equals(newValue)))
      {
        dataBool.put(attrKey, newValue);
        markBoolDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetNumberAttr ( String attrKey, java.math.BigDecimal newValue )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
     try {
      checkGenericNumberLoad();
      java.math.BigDecimal oldValue = (java.math.BigDecimal)dataNumber.get(attrKey);
      if ( (oldValue!=null||newValue!=null) && (oldValue==null||!oldValue.equals(newValue)))
      {
        dataNumber.put(attrKey, newValue);
        markNumberDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetStringAttr ( String attrKey, String newValue )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
     try {
      checkGenericStringLoad();
      String oldValue = (String)dataString.get(attrKey);
      if ( (oldValue!=null||newValue!=null) && (oldValue==null||!oldValue.equals(newValue)))
      {
        dataString.put(attrKey, newValue);
        markStringDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetBinaryAttr ( String attrKey, byte[] newValue )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
     try {
      checkGenericBinaryLoad();
      byte[] oldValue = (byte[])dataBinary.get(attrKey);
      if ( (oldValue!=null||newValue!=null) && (oldValue==null||!oldValue.equals(newValue)))
      {
        dataBinary.put(attrKey, newValue);
        markBinaryDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Set a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @param newValue The value which is written to the DO
   */
   public void genericSetObjectAttr ( String attrKey, GenericStoreDO newValue )
   throws DataObjectException, GenericStoreDefinitionException {
	checkTypeDefinition(this.getClass().getName());
	checkAttrDefinition(attrKey);
     try {
      checkGenericObjectLoad();
      GenericStoreDO oldValue = (GenericStoreDO)dataObject.get(attrKey);
      if ( (oldValue!=null||newValue!=null) && (oldValue==null||!oldValue.equals(newValue)))
      {
        dataObject.put(attrKey, newValue.getOId());
        markObjectDirty(attrKey);
      }
    } catch (Exception e) {
      throw new DataObjectException("GenericSetAttr error",e);
    }
   }
  /**
   * Load the object extension
   *
   */
  public void loadGenericObjectDO()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	ObjectId id = getOId();
	if ( null == id ) return;
	if ( ! isPersistent() ) return;
	GenericObjectQuery query = new GenericObjectQuery ();

	query.setQueryEXTERNALOID( id.toBigDecimal() );
	query.requireUniqueInstance();
	try {
	    objectExtension = query.getNextDO();
	} catch ( NonUniqueQueryException e ) {
	    throw new ObjectIdException( "Duplicate ObjectId" );
	}
  }
  /**
   * Load the enum attributes
   *
   */
  public void loadGenericEnumData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataEnum ) dataEnum = new TreeMap();
	if ( null == typeEnum ) typeEnum = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeEnumValueQuery query = new AttributeEnumValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeEnumValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            dataEnum.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),tempArray[x].getENUMVALUEDEFINITIONOID().getKEYVALUE());
            typeEnum.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),tempArray[x].getENUMVALUEDEFINITIONOID().getENUMTYPEDEFINITIONOID().getKEYVALUE());
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Load the date attributes
   *
   */
  public void loadGenericDateData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataDate ) dataDate = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeDateValueQuery query = new AttributeDateValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeDateValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            dataDate.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),tempArray[x].getDATEVALUE());
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Load the date attributes
   *
   */
  public void loadGenericBoolData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataBool ) dataBool = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeBoolValueQuery query = new AttributeBoolValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeBoolValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            dataBool.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),new Boolean(tempArray[x].getBOOLVALUE()));
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Load the number attributes
   *
   */
  public void loadGenericNumberData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataNumber ) dataNumber = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeNumberValueQuery query = new AttributeNumberValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeNumberValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            AttributeDefinitionDO attrdef = tempArray[x].getATTRIBUTEDEFINITIONOID();
            java.math.BigDecimal scale = attrdef.getATTRSCALE();
            java.math.BigDecimal newval = tempArray[x].getNUMBERVALUE().movePointLeft(scale.intValue());
            dataNumber.put(attrdef.getKEYVALUE(),newval);
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Load the string attributes
   *
   */
  public void loadGenericStringData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataString ) dataString = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeStringValueQuery query = new AttributeStringValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeStringValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            dataString.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),tempArray[x].getSTRINGVALUE());
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Load the string attributes
   *
   */
  public void loadGenericBinaryData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataBinary ) dataBinary = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeBinaryValueQuery query = new AttributeBinaryValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeBinaryValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            dataBinary.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),tempArray[x].getBINARYVALUE());
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Load the pointer attributes
   *
   */
  public void loadGenericObjectData()
  throws SQLException, ObjectIdException, DataObjectException, QueryException
  {
	if ( null == dataObject ) dataObject = new TreeMap();
	if ( objectExtension == null ) return;
	AttributeObjectValueQuery query = new AttributeObjectValueQuery ();
	query.setQueryGENERICOBJECTOID(objectExtension);
	AttributeObjectValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            dataObject.put(tempArray[x].getATTRIBUTEDEFINITIONOID().getKEYVALUE(),tempArray[x].getOBJECTVALUE());
          }
	} catch ( NonUniqueQueryException e ) {
	}
  }
  /**
   * Read a certain attribute definition
   *
   * @param attrKey The name of the attribute definition
   */
  public static AttributeDefinitionDO findAttrDef(String attrKey)
  throws DataObjectException
  {
	AttributeDefinitionQuery query = new AttributeDefinitionQuery ();
        AttributeDefinitionDO attrDef = null;
	try {
  	  query.setQueryKEYVALUE(attrKey);
	  query.requireUniqueInstance();
	  attrDef = query.getNextDO();
	} catch ( Exception e ) {
	    throw new DataObjectException( "AttributeDefinition error",e );
	}
        return attrDef;
  }
  /**
   * Read a certain object-type definition
   *
   * @param attrKey The name of the type definition
   */
  public static TypeDefinitionDO findTypeDef(String typeKey)
  throws DataObjectException
  {
	TypeDefinitionQuery query = new TypeDefinitionQuery ();
        TypeDefinitionDO typeDef = null;
	try {
  	  query.setQueryKEYVALUE(typeKey);
	  query.requireUniqueInstance();
	  typeDef = query.getNextDO();
	} catch ( Exception e ) {
	    throw new DataObjectException( "TypeDefinition error",e );
	}
        return typeDef;
  }
  /**
   * Check if state is a start-state
   *
   * @param startTypeKey The key of the state-machine
   * @param startValueKey The key of the state-value
   * @result Whether state is a start-state
   */
  public static boolean genericIsBeginState(String startTypeKey, String startValueKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkEnumValueDefinition(startTypeKey, startValueKey);
	return findBeginStates(startTypeKey).contains(startValueKey);
  }
  /**
   * Check if state is an end-state
   *
   * @param endTypeKey The key of the state-machine
   * @param endValueKey The key of the state-value
   * @result Whether state is an end-state
   */
  public static boolean genericIsEndState(String endTypeKey, String endValueKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkEnumValueDefinition(endTypeKey, endValueKey);
	return findEndStates(endTypeKey).contains(endValueKey);
  }
  /**
   * Find the start-states for a certain statemachine (EnumTypeDef)
   *
   * @param typeKey The keyvalue of the enum type definition (statemachine)
   * @result Vector containing all key-values of start-states (EnumValueDef's)
   */
  public static Vector findBeginStates(String enumTypeKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkEnumTypeDefinition(enumTypeKey);
	EnumTypeDefinitionDO typeDef = findEnumTypeDef(enumTypeKey);
	EnumValueDefinitionQuery query = new EnumValueDefinitionQuery ();
	EnumValueDefinitionDO[] arrValueDef = null;
	try {
  	  query.setQueryENUMTYPEDEFINITIONOID(typeDef);
  	  query.setQueryBEGINSTATE(true);
	  arrValueDef = query.getDOArray();
	} catch ( Exception e ) {
	    throw new DataObjectException( "findBeginStates error",e );
	}
	Vector result = new Vector();
	for(int x=0;x<arrValueDef.length;x++)
		result.add(arrValueDef[x].getKEYVALUE());
	return result;
  }
  /**
   * Find the end-states for a certain statemachine (EnumTypeDef)
   *
   * @param typeKey The keyvalue of the enum type definition (statemachine)
   * @result Vector containing all key-values of start-states (EnumValueDef's)
   */
  public static Vector findEndStates(String enumTypeKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkEnumTypeDefinition(enumTypeKey);
	EnumTypeDefinitionDO typeDef = findEnumTypeDef(enumTypeKey);
	EnumValueDefinitionQuery query = new EnumValueDefinitionQuery ();
	EnumValueDefinitionDO[] arrValueDef = null;
	try {
  	  query.setQueryENUMTYPEDEFINITIONOID(typeDef);
  	  query.setQueryENDSTATE(true);
	  arrValueDef = query.getDOArray();
	} catch ( Exception e ) {
	    throw new DataObjectException( "findEndStates error",e );
	}
	Vector result = new Vector();
	for(int x=0;x<arrValueDef.length;x++)
		result.add(arrValueDef[x].getKEYVALUE());
	return result;
  }
  /**
   * Find a defined state-transition
   *
   * @param currentStateKey The keyvalue of the current state (EnumValueDefinition)
   * @param currentTypeKey The keyvalue of the state-machine (EnumTypeDefinition)
   * @param newStateKey The keyvalue of the new state (EnumValueDefinition)
   * @result The StateTransition if it exists, else NULL
   */
  public static StateTransitionDO findStateTransition(String currentTypeKey, String currentStateKey, String newStateKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkEnumValueDefinition(currentTypeKey, currentStateKey);
	checkEnumValueDefinition(currentTypeKey, newStateKey);
	EnumValueDefinitionDO currentValueDef = findEnumValueDef(currentTypeKey, currentStateKey);
	EnumValueDefinitionDO newValueDef = findEnumValueDef(currentTypeKey, newStateKey);
	StateTransitionQuery query = new StateTransitionQuery ();
	StateTransitionDO result = null;
	try {
  	  query.setQueryENUMVALUEDEFINITIONFROMOID(currentValueDef);
  	  query.setQueryENUMVALUEDEFINITIONTOOID(newValueDef);
	  query.requireUniqueInstance();
	  result = query.getNextDO();
	} catch ( Exception e ) {
	    throw new DataObjectException( "findEndStates error",e );
	}
	return result;
  }
  /**
   * Find a defined state-transition
   *
   * @param attrKey The keyvalue of the attribute
   * @param newStateKey The keyvalue of the new state (EnumValueDefinition)
   * @result The StateTransition if it exists, else NULL
   */
  public StateTransitionDO genericGetStateTransition(String attrKey, String newStateKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkAttrDefinition(attrKey);
	return findStateTransition(genericGetEnumAttr(attrKey), genericGetEnumTypeAttr(attrKey), newStateKey);
  }
  /**
   * Find allowed state-transitions
   *
   * @param stateKey The keyvalue of the source state (EnumValueDefinition)
   * @param typeKey The keyvalue of the state-machine (EnumTypeDefinition)
   * @result Vector containing all keys of allowed target-states (EnumValueDefinition's)
   */
  public static Vector findAllowedTargetStates(String typeKey, String stateKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkEnumValueDefinition(typeKey, stateKey);
	EnumValueDefinitionDO valueDef = findEnumValueDef( typeKey, stateKey);
	Vector result = new Vector();
	try {
		StateTransitionDO[] trans = valueDef.getStateTransitionDOArray_ENUMVALUEDEFINITIONFROMOID();
      	for(int x=0;x<trans.length;x++)
			result.add(trans[x].getENUMVALUEDEFINITIONTOOID().getKEYVALUE());
	} catch ( Exception e ) {
	    throw new DataObjectException( "findAllowedTargetStates error",e );
	}
	return result;
  }
  /**
   * Check if state-transition is allowed
   *
   * @param attrKey The keyvalue of the Attribute
   * @param targetStateKey The keyvalue of the state (EnumValueDefinition) which should be checked based on the current state
   * @result true if allowed, false if not
   */
  public boolean genericGetTargetStateAllowed(String attrKey, String targetStateKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkAttrDefinition(attrKey);
	return (findAllowedTargetStates(genericGetEnumTypeAttr(attrKey), genericGetEnumAttr(attrKey))).contains(targetStateKey);
  }
  /**
   * Find allowed state-transitions
   *
   * @param attrKey The keyvalue of the Attribute
   * @result Vector containing all keys of allowed target-states (EnumValueDefinition's) based on the current state
   */
  public Vector genericGetAllowedTargetStates(String attrKey)
  throws DataObjectException, GenericStoreDefinitionException
  {
	checkAttrDefinition(attrKey);
	return findAllowedTargetStates(genericGetEnumTypeAttr(attrKey), genericGetEnumAttr(attrKey));
  }
  /**
   * Find a certain enum-type definition
   *
   * @param typeKey The name of the type definition
   */
  public static EnumTypeDefinitionDO findEnumTypeDef(String typeKey)
  throws DataObjectException
  {
	EnumTypeDefinitionQuery query = new EnumTypeDefinitionQuery ();
        EnumTypeDefinitionDO typeDef = null;
	try {
  	  query.setQueryKEYVALUE(typeKey);
	  query.requireUniqueInstance();
	  typeDef = query.getNextDO();
	} catch ( Exception e ) {
	    throw new DataObjectException( "EnumTypeDefinition error",e );
	}
        return typeDef;
  }
  /**
   * find a certain enum-value definition
   *
   * @param typeKey The name of the type definition
   * @param valueKey The name of the type definition
   */
  public static EnumValueDefinitionDO findEnumValueDef(String typeKey, String valueKey)
  throws DataObjectException
  {
	EnumValueDefinitionQuery query = new EnumValueDefinitionQuery ();
        EnumValueDefinitionDO valueDef = null;
	try {
  	  query.setQueryKEYVALUE(valueKey);
  	  query.setQueryENUMTYPEDEFINITIONOID(findEnumTypeDef(typeKey));
	  query.requireUniqueInstance();
	  valueDef = query.getNextDO();
	} catch ( Exception e ) {
	    throw new DataObjectException( "EnumTypeDefinition error",e );
	}
        return valueDef;
  }


  /**
   * Read a certain generic attribute
   *
   * @param attrKey The name of the attribute which will be read
   * @return The desired generic attribute value of the DO
   */
   public GenericStoreDO genericGetObjectAttr (String attrKey)
   throws DataObjectException, GenericStoreDefinitionException {
	checkAttrDefinition(attrKey);
      checkGenericObjectLoad();
      java.math.BigDecimal bd = (java.math.BigDecimal) dataObject.get(attrKey);
      AttributeDefinitionDO attrdef = findAttrDef(attrKey);
      String clsName = attrdef.getATTROBJECTCLASS();
      try {
        Class clsGSDO = Class.forName(clsName);
        GenericStoreDO gsDO = (GenericStoreDO)clsGSDO.newInstance();
        return gsDO.createForExisting(bd);
      } catch (Exception e)
      {
	throw new DataObjectException(e.getMessage(),e);
      }
   }
  /**
   * Write the generic attributes to the DB
   * Must be called at the end of subclass save-method which an explicit transaction
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  public void save(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (needsGenericSaveData())
    {
      if (objectExtension == null)
      {
         String clsName = this.getClass().getName();
         if (!clsName.equals("org.webdocwf.util.genericstore.data.GenericObjectDO"))
         {
           try {
           objectExtension = GenericObjectDO.createVirgin();
           } catch (Exception e)
           {
              throw new DataObjectException("Create GenericObjectDO error",e);
           }
         }
         else
         {
            objectExtension = (GenericObjectDO)this;
         }
         TypeDefinitionDO typeDef = findTypeDef(clsName);
         objectExtension.setTYPEDEFINITIONOID(typeDef);
         objectExtension.setEXTERNALOID(this.getOId().toBigDecimal());
         objectExtension.save(dbt);
      }
      saveGeneric(dbt);
    }
  }
  /**
   * Write the generic attributes to the DB
   *
   */
  public void save()
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    DBTransaction dbt = DODS.getDatabaseManager().createTransaction();

    save(dbt);
    dbt.commit();
    dbt.release();
  }
  /**
   * Delete all generic attributes together with the DO
   * Must be called at the beginning of the subclass-delete methid with an explicit transaction
   *
   * @param dbt The transaction in which the delete is performed in the DB
   */
  public void delete(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    checkGenericObjectDOLoad();
    if (objectExtension != null)
    {
      objectExtension.delete(dbt);
      genericDeleteReferences(dbt);
    }
  }
  /**
   * Delete all generic pointers to a DO
   *
   * @param dbt The transaction in which the delete is performed in the DB
   */
  private void genericDeleteReferences(DBTransaction dbt)
  throws SQLException, DataObjectException, QueryException
  {
	AttributeObjectValueQuery query = new AttributeObjectValueQuery ();
	query.setQueryOBJECTVALUE(this.getOId().toBigDecimal());
	AttributeObjectValueDO[] tempArray;
	try {
          tempArray = query.getDOArray();
          for (int x=0;x<tempArray.length;x++)
          {
            tempArray[x].delete(dbt);
          }
	} catch ( Exception e ) {
          throw new DataObjectException("genericDeleteReferences error",e);
	}
  }
  /**
   * Delete all generic attributes
   *
   */
  public void delete()
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    checkGenericObjectDOLoad();
    if (objectExtension != null)
    {
      objectExtension.delete();
    }
  }
  /**
   * Save generic attribute changes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGeneric(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    saveGenericDates(dbt);
    saveGenericBools(dbt);
    saveGenericNumbers(dbt);
    saveGenericStrings(dbt);
    saveGenericObjects(dbt);
    saveGenericEnums(dbt);
    saveGenericBinary(dbt);
    dirtyDates = null;
    dirtyBools = null;
    dirtyNumbers = null;
    dirtyStrings = null;
    dirtyObjects = null;
    dirtyEnums = null;
    dirtyBinarys = null;
  }
  /**
   * Check if object extensions need to be saved
   *
   * @return Whether object extensions need to be saved
   */
  private boolean needsGenericSaveData()
  {
    return ((objectExtension != null)||hasGenericDataValues());
  }
  /**
   * Check if object extension attributes need to be saved
   *
   * @return Whether object extension attributes need to be saved
   */
  private boolean hasGenericDataValues()
  {
    return hasGenericValues(dataDate)||hasGenericValues(dataNumber)||hasGenericValues(dataString)||hasGenericValues(dataObject)||hasGenericValues(dataEnum)||hasGenericValues(dataBool);
  }
  /**
   * Check if object extension attributes need to be saved
   *
   * @return Whether object extension attributes need to be saved
   */
  private boolean hasGenericValues(TreeMap map)
  {
    boolean ret = false;
    if (map != null)
    {
      Object vals[] = map.values().toArray();
      for(int x=0;x<vals.length;x++)
      {
        if (vals[x]!=null) ret = true;
      }
    }
    return ret;
  }
  /**
   * Get all generic enum attributes types
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllEnumTypeAttrs()
  throws DataObjectException
  {
    checkGenericEnumLoad();
    return (TreeMap)typeEnum.clone();
  }
  /**
   * Get all generic enum attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllEnumAttrs()
  throws DataObjectException
  {
    checkGenericEnumLoad();
    return (TreeMap)dataEnum.clone();
  }
  /**
   * Get all generic date attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllDateAttrs()
  throws DataObjectException
  {
    checkGenericDateLoad();
    return (TreeMap)dataDate.clone();
  }
  /**
   * Get all generic bool attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllBoolAttrs()
  throws DataObjectException
  {
    checkGenericBoolLoad();
    return (TreeMap)dataBool.clone();
  }
  /**
   * Get all generic number attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllNumberAttrs()
  throws DataObjectException
  {
    checkGenericNumberLoad();
    return (TreeMap)dataNumber.clone();
  }
  /**
   * Get all generic string attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllStringAttrs()
  throws DataObjectException
  {
    checkGenericStringLoad();
    return (TreeMap)dataString.clone();
  }
  /**
   * Get all generic string attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllBinaryAttrs()
  throws DataObjectException
  {
    checkGenericBinaryLoad();
    return (TreeMap)dataBinary.clone();
  }
  /**
   * Get all generic pointer attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllObjectAttrs()
  throws DataObjectException
  {
    checkGenericObjectLoad();
    return (TreeMap)dataObject.clone();
  }



  /**
   * Get all generic enum attributes types
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllEnumTypeAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllEnumTypeAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetEnumAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic enum attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllEnumAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllEnumAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetEnumAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic date attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllDateAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllDateAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetDateAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic bool attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllBoolAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllBoolAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetBoolAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic number attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllNumberAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllNumberAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetNumberAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic string attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllStringAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllStringAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetStringAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic string attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllByteAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllBinaryAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetBinaryAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }
  /**
   * Get all generic pointer attributes
   *
   * @return All generic name/value pairs
   */
  public TreeMap genericGetAllObjectAttrs(User usr)
  throws DataObjectException, AccessEvalException, GenericStoreDefinitionException
  {
	TreeMap dataTree = genericGetAllObjectAttrs();
        TreeMap result = new TreeMap();
	Iterator itData = dataTree.keySet().iterator();
	while (itData.hasNext())
	{
		String attrKey = (String)itData.next();
		if (hasDOGenericGetObjectAttrAccess( attrKey, usr ))
                        result.put(attrKey, dataTree.get(attrKey));
	}
	return result;
  }



  private TreeSet dirtyEnums = null;
  /**
   * Mark a specific attribute as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markEnumDirty(String attrKey)
  {
    if (dirtyEnums == null) dirtyEnums = new TreeSet();
    dirtyEnums.add(attrKey);
  }
  private TreeSet dirtyDates = null;
  /**
   * Mark a specific attribute as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markDateDirty(String attrKey)
  {
    if (dirtyDates == null) dirtyDates = new TreeSet();
    dirtyDates.add(attrKey);
  }
  private TreeSet dirtyBools = null;
  /**
   * Mark a specific attribute as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markBoolDirty(String attrKey)
  {
    if (dirtyBools == null) dirtyBools = new TreeSet();
    dirtyBools.add(attrKey);
  }
  private TreeSet dirtyNumbers = null;
  /**
   * Mark a specific attrubte as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markNumberDirty(String attrKey)
  {
    if (dirtyNumbers == null) dirtyNumbers = new TreeSet();
    dirtyNumbers.add(attrKey);
  }
  private TreeSet dirtyStrings = null;
  /**
   * Mark a specific attrubte as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markStringDirty(String attrKey)
  {
    if (dirtyStrings == null) dirtyStrings = new TreeSet();
    dirtyStrings.add(attrKey);
  }
  private TreeSet dirtyBinarys = null;
  /**
   * Mark a specific attrubte as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markBinaryDirty(String attrKey)
  {
    if (dirtyBinarys == null) dirtyBinarys = new TreeSet();
    dirtyBinarys.add(attrKey);
  }
  private TreeSet dirtyObjects = null;
  /**
   * Mark a specific attrubte as dirty
   *
   * @param attrKey The name of the attribute
   */
  private void markObjectDirty(String attrKey)
  {
    if (dirtyObjects == null) dirtyObjects = new TreeSet();
    dirtyObjects.add(attrKey);
  }
  /**
   * Save all enum attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericEnums(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyEnums != null)
      {
      Iterator itDirty = dirtyEnums.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericEnum( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save all date attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericDates(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyDates != null)
      {
      Iterator itDirty = dirtyDates.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericDate( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save all bool attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericBools(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyBools != null)
      {
      Iterator itDirty = dirtyBools.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericBool( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save all number attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericNumbers(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyNumbers != null)
      {
      Iterator itDirty = dirtyNumbers.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericNumber( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save all string attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericStrings(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyStrings != null)
      {
      Iterator itDirty = dirtyStrings.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericString( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save all binary attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericBinary(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyBinarys != null)
      {
      Iterator itDirty = dirtyBinarys.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericBinary( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save all pointer attributes
   *
   * @param dbt The transaction in which the changes are written to the DB
   */
  private void saveGenericObjects(DBTransaction dbt)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    if (dirtyObjects != null)
      {
      Iterator itDirty = dirtyObjects.iterator();
      while ( itDirty.hasNext() )
          {
          saveGenericObject( dbt, (String) itDirty.next() );
          }
      }
  }
  /**
   * Save an enum attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericEnum(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeEnumValueQuery query = new AttributeEnumValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeEnumValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    String dbval = null;
    if (dbDO != null) dbval = dbDO.getENUMVALUEDEFINITIONOID().getKEYVALUE();
    String newval = (String)dataEnum.get(attrKey);
    String newtype = (String)typeEnum.get(attrKey);
    if (newval != null)
    {
      if (dbval != null)
      {
        //update
        dbDO.setENUMVALUEDEFINITIONOID(getEnumValueDefinition(attrdef, newtype, newval));
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeEnumValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setENUMVALUEDEFINITIONOID(getEnumValueDefinition(attrdef, newtype, newval));
        dbDO.save(dbt);
      }
    }
    else
    {
      if (dbval != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }
  /**
   * Find an EnumValueDefinition for a certain AttributeDefinition
   *
   * @param attrdef The AttributeDefinition
   * @param valueKey The key of the EnumValueDefinition
   * @param typeKey The key of the EnumTypeDefinition, can be null (means take existing or attr-default)
   */
  private EnumValueDefinitionDO getEnumValueDefinition(AttributeDefinitionDO attrdef, String typeKey, String valueKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    EnumTypeDefinitionDO enumtypedef = null;
    if (typeKey != null) enumtypedef = findEnumTypeDef(typeKey);
    if (enumtypedef == null) enumtypedef=attrdef.getENUMTYPEDEFINITIONOID();
    EnumValueDefinitionQuery query = new EnumValueDefinitionQuery();
    query.setQueryENUMTYPEDEFINITIONOID(enumtypedef);
    query.setQueryKEYVALUE(valueKey);
    query.requireUniqueInstance();
    EnumValueDefinitionDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    if (dbDO == null) throw new DataObjectException("Enum Value not found ! attrKey="+attrdef.getKEYVALUE()+", valueKey="+valueKey);
    return dbDO;
  }
  /**
   * Save a date attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericDate(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeDateValueQuery query = new AttributeDateValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeDateValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    java.sql.Timestamp dbval = null;
    if (dbDO != null) dbval = dbDO.getDATEVALUE();
    java.sql.Timestamp newval = (java.sql.Timestamp)dataDate.get(attrKey);
    if (newval != null)
    {
      if (dbval != null)
      {
        //update
        dbDO.setDATEVALUE(newval);
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeDateValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setDATEVALUE(newval);
        dbDO.save(dbt);
      }
    }
    else
    {
      if (dbval != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }
  /**
   * Save a bool attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericBool(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeBoolValueQuery query = new AttributeBoolValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeBoolValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    Boolean dbval = null;
    if (dbDO != null) dbval = new Boolean(dbDO.getBOOLVALUE());
    Boolean newval = (Boolean)dataBool.get(attrKey);
    if (newval != null)
    {
      if (dbval != null)
      {
        //update
        dbDO.setBOOLVALUE(newval.booleanValue());
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeBoolValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setBOOLVALUE(newval.booleanValue());
        dbDO.save(dbt);
      }
    }
    else
    {
      if (dbval != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }
  /**
   * Save a string attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericString(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeStringValueQuery query = new AttributeStringValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeStringValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    String dbval = null;
    if (dbDO != null) dbval = dbDO.getSTRINGVALUE();
    String newval = (String)dataString.get(attrKey);
    if (newval != null)
    {
      if (dbval != null)
      {
        //update
        dbDO.setSTRINGVALUE(newval);
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeStringValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setSTRINGVALUE(newval);
        dbDO.save(dbt);
      }
    }
    else
    {
      if (dbval != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }
  /**
   * Save a binary attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericBinary(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeBinaryValueQuery query = new AttributeBinaryValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeBinaryValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    byte[] dbval = null;
    if (dbDO != null) dbval = dbDO.getBINARYVALUE();
    byte[] newval = (byte[])dataBinary.get(attrKey);
    if (newval != null)
    {
      //if (dbval != null) <-- original Line
      // --> Patch for JTurbo; returns null if attribute is 0x in database
      // --> if 0x is stored in database inserts instead of updates will be executed
      // --> proceed update if DO exists
      if (dbDO != null)
      {
        //update
        dbDO.setBINARYVALUE(newval);
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeBinaryValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setBINARYVALUE(newval);
        dbDO.save(dbt);
      }
    }
    else
    {
      //if (dbval != null) <-- original Line
      // --> Patch for JTurbo; returns null if attribute is 0x in database
      // --> if 0x is stored in database inserts instead of updates will be executed
      // --> proceed update if DO exists
      if (dbDO != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }

  /**
   * Save a pointer attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericObject(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeObjectValueQuery query = new AttributeObjectValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeObjectValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    java.math.BigDecimal dbval = null;
    if (dbDO != null) dbval = dbDO.getOBJECTVALUE();
    java.math.BigDecimal newval = (java.math.BigDecimal)dataObject.get(attrKey);
    if (newval != null)
    {
      if (dbval != null)
      {
        //update
        dbDO.setOBJECTVALUE(newval);
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeObjectValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setOBJECTVALUE(newval);
        dbDO.save(dbt);
      }
    }
    else
    {
      if (dbval != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }
  /**
   * Save a number attribute
   *
   * @param dbt The transaction in which the changes are written to the DB
   * @param attrKey The name of the attribute
   */
  private void saveGenericNumber(DBTransaction dbt, String attrKey)
  throws SQLException, DatabaseManagerException, DataObjectException, RefAssertionException, DBRowUpdateException, QueryException {
    AttributeNumberValueQuery query = new AttributeNumberValueQuery();
    AttributeDefinitionDO attrdef = findAttrDef(attrKey);
    query.setQueryATTRIBUTEDEFINITIONOID(attrdef);
    query.setQueryGENERICOBJECTOID(objectExtension);
    query.requireUniqueInstance();
    AttributeNumberValueDO dbDO = null;
    try {
    dbDO = query.getNextDO(); }
    catch (NonUniqueQueryException e)
    {
      throw new DataObjectException("AttributeDefinition not unique",e);
    }
    java.math.BigDecimal dbval = null;
    if (dbDO != null) dbval = dbDO.getNUMBERVALUE();
    java.math.BigDecimal newval = (java.math.BigDecimal)dataNumber.get(attrKey);
    java.math.BigDecimal scale = attrdef.getATTRSCALE();
    newval = newval.movePointRight(scale.intValue());
    if (newval != null)
    {
      if (dbval != null)
      {
        //update
        dbDO.setNUMBERVALUE(newval);
        dbDO.save(dbt);
      }
      else
      {
        //insert
        try {
        dbDO = AttributeNumberValueDO.createVirgin();}
        catch (ObjectIdException e)
        {
          throw new DataObjectException("ObjectIdException",e);
        }
        dbDO.setATTRIBUTEDEFINITIONOID(attrdef);
        dbDO.setGENERICOBJECTOID(objectExtension);
        dbDO.setNUMBERVALUE(newval);
        dbDO.save(dbt);
      }
    }
    else
    {
      if (dbval != null)
      {
        //delete
        dbDO.delete(dbt);
      }
    }
  }
  /**
   * Check whether a certain Attribute-Definition exists
   *
   * @param attrKey The key of the attribute definition
   */
  static public void checkAttrDefinition(String attrKey)
  throws DataObjectException, GenericStoreDefinitionException {
	AttributeDefinitionDO attrDef=null;
	attrDef = findAttrDef(attrKey);
	if (attrDef==null)
	{
		throw new GenericStoreDefinitionException("AttributeDefinition " + attrKey + " not defined in GenericStore !");
	}
  }
  /**
   * Check whether a certain EnumValue-Definition exists
   *
   * @param typeKey The type of the EnumValue
   * @param valueKey The key of the EnumValue
   */
  static public void checkEnumValueDefinition(String typeKey, String valueKey)
  throws DataObjectException, GenericStoreDefinitionException {
	checkEnumTypeDefinition(typeKey);
	EnumValueDefinitionDO enumValueDef=null;
	enumValueDef = findEnumValueDef(typeKey, valueKey);
	if (enumValueDef==null)
	{
		throw new GenericStoreDefinitionException("EnumValueDefinition "+valueKey+" in EnumType "+typeKey+" not defined in GenericStore !");
	}
  }
  /**
   * Check whether a certain EnumType-Definition exists
   *
   * @param typeKey The key of the EnumType definition
   */
  static public void checkEnumTypeDefinition(String typeKey)
  throws DataObjectException, GenericStoreDefinitionException {
	EnumTypeDefinitionDO enumTypeDef=null;
	enumTypeDef = findEnumTypeDef(typeKey);
	if (enumTypeDef==null)
	{
		throw new GenericStoreDefinitionException("EnumTypeDefinition "+typeKey+" not defined in GenericStore !");
	}
  }
  /**
   * Check whether a certain Type-Definition exists
   *
   * @param typeKey The key of the Type definition
   */
  static public void checkTypeDefinition(String key)
  throws DataObjectException, GenericStoreDefinitionException {
	TypeDefinitionDO typeDef=null;
	typeDef = findTypeDef(key);
	if (typeDef==null)
	{
		throw new GenericStoreDefinitionException("TypeDefinition "+key+" not defined in GenericStore !");
	}
  }
}